"""
Version : 1.0 (06-18-2022).

DEPENDENCIES:
     'lambert_W.py' in the folder 'utils'

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np
from proxop.utils.lambert_W import lambert_W


class EntropyL0:
    r"""Compute the proximity operator and the evaluation of the gamma*f.

      where the function f is defined as:

                                       /   x*log(x) + w      if x>0
               f(x) = min{ x^2, w}  = |    0                 if x=0
                                      \    +inf             otherwise

    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise SUM.

     INPUTS
    ========
    x     - ND array
    w     - positive, scalar or ND array with the same size as 'x'
    gamma - positive, scalar or ND array with the same size as 'x'[default: gamma=1]


    Note: When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>EntropyL0(w=2)(x) will
    return a scalar even if x is a vector:

    >>> EntropyL0(w=2)(np.array([1, 2, 3]))
    10.68213122712422

    But as expected, >>>EntropyL0(w=2).prox(x)
    will return a vector with the same size as x:

    >>> EntropyL0(w=2).prox(np.array([1, 2, 3]))
    array([0.       , 0.       , 1.5571456])
    """

    def __init__(
            self,
            w: Union[np.ndarray, float]
    ):

        if np.any(w <= 0):
            raise ValueError("'w' (or all of its elements) must be strictly positive")
        self.w = w

    def prox(self, x: np.ndarray, gamma: Union[np.ndarray, float] = 1.0) -> np.ndarray:
        self._check(x, gamma)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))

        prox_x = np.zeros(np.shape(x))
        # 3rd branch
        # we use a taylor approximation to avoid divergence when the input is too big
        u = np.log(1 / gamma) + x / gamma - 1
        mask = u > 100

        prox_x[mask] = u[mask] - u[mask] / (1 + u[mask]) * np.log(u[mask])

        # we use the Lambert_W function otherwise
        mask = np.logical_not(mask)
        gamma_m = 1.0*gamma
        if np.size(gamma) > 1:
            gamma_m = gamma[mask]
        prox_x[mask] = gamma_m * lambert_W(1 / gamma_m * np.exp(x[mask] / gamma_m - 1))
        mask = prox_x**2 + 2 * gamma * prox_x < 2 * self.w * gamma
        # 1st and 2nd branch
        prox_x[mask] = 0
        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        if np.any(x < 0):
            return np.inf
        mask = x > 0
        w = self.w
        if np.size(w) > 1:
            w = w[mask]
        fun_x = x[mask] * np.log(x[mask]) + w
        return np.sum(fun_x)

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its elements) must be strictly positive"
            )
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise ValueError("'gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 ( 06-21-2022).

DEPENDENCIES:
     - 'AbsValue.py' located in the folder 'scalar'
     - 'prox_svd.py' located in the folder 'utils'
     - 'NuclearNorm.py' located in the folder 'multi'
     - 'Linf.py' located in the folder 'multi'
     - 'Max.py' in the folder 'multi'
     - 'L1Ball.py' in the folder 'indicator'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np
from proxop.indicator.L1Ball import L1Ball
from proxop.utils.prox_svd import prox_svd
from proxop.multi.NuclearNorm import NuclearNorm


class NuclearBall:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the indicator of the constraint set:

                         ||X||_N =||s||_1 <= eta   (nuclear norm ball constraint)

            where X = U*diag(s)*V.T \in R^{M*N}

     INPUTS
    ========
     x         -  (M,N) -array_like ( representing an M*N matrix )
     gamma     - positive, scalar or ND array compatible with the size of 'x'
                 [default: gamma=1]
     eta       - positive, scalar or ND array compatible with the size of 'x'
    """

    def __init__(
            self,
            eta: Union[float, np.ndarray]
    ):
        if np.any(eta <= 0):
            ValueError("'eta' must be strictly positive")
        self.eta = eta

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1) -> np.ndarray:
        return prox_svd(x, 1, L1Ball(eta=gamma).prox)

    def __call__(self, x: np.ndarray) -> float:
        nuclear_norm = NuclearNorm()(x)
        if nuclear_norm <= self.eta:
            return 0
        return np.inf



"""
Version : 1.0 (06-07-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class Square:
    r"""Computes the proximity operator and the evaluation of gamma*f.

    Where f is the 'Square' function defined as:


            f(x) = 1/2 * x^2

    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise :

    -When calling the function, the output is a scalar (sum of the element-wise
    results ) .

    - But for the proximity operator (method 'prox'), the output has the same shape
    as the input 'x'.

     INPUTS
    ========
     x     - ND array
     gamma - positive, scalar or ND array with the same size as 'x' (default: gamma=1)

    =======
    Examples
    ========

     Evaluate the function  f:

     >>> Square()( 4 )
     8.0

      Compute the result element-wise sum for vector inputs :

     >>> Square()([-1, 3, 2] )
     7.0

     Compute the proximity operator at a given point :

     >>> Square().prox(  [-2, 3, 4 ])
     array([-1. ,  1.5,  2. ])

     Use a scale factor 'gamma'>0 to compute the proximity operator of
     the function 'gamma*f'

     >>> Square().prox( [-2, 3, 4, 6 ], gamma=2)
     array([-0.66666667,  1.        ,  1.33333333,  2.        ])
    """

    def __init__(self):
        pass

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        self._check(x, gamma)
        return x / (1 + gamma)

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        return np.sum(0.5 * x**2)

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its components if it is an array) must "
                + "be strictly positive"
            )
        if (np.size(gamma) > 1) and (np.shape(gamma) != np.shape(x)):
            raise ValueError("gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 (06-18-2022).

DEPENDENCIES:
     'newton.py' in the folder 'utils'

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np
from proxop.utils.newton import newton_


class Root:
    r"""Compute the proximity operator and the evaluation of the gamma*f.

      where the function f is defined as:


                  f(x) = |x|^q  with  0 < q < 1


    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
    x     - ND array
    gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    q     - in ]0, 1[ , scalar or ND array with the same size as 'x'

    Note: When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>Root(q=0.25)(x) will
    return a scalar even if x is a vector:

    >>> Root(q=0.25)(np.array([-1, 2, 3]))
    3.50528112795521

    But as expected, >>>Root(q=0.25).prox(x)
    will return a vector with the same size as x:

    >>> Root(q=0.25).prox(np.array([-1, 2, 3]))
    array([0.        , 1.8418772 , 2.88712686])
    """

    def __init__(
            self,
            q: Union[float, np.ndarray]
    ):

        if np.any(q <= 0) or np.any(q >= 1):
            raise ValueError(
                "'q' (or all of its components if it"
                + " is an array) must belong to ]0, 1["
            )
        self.q = q

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        self._check(x, gamma)
        q = self.q

        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        eps = 1.0e-16
        # select the branches
        alpha = gamma * (np.abs(x) + eps) ** (q - 2)
        mask = alpha <= (2 * (1 - q) / (2 - q)) ** (1 - q) / (2 - q)
        alpha = alpha[mask]
        if np.size(q) > 1:
            q = q[mask]

        # 1st branch
        prox_x = np.zeros(np.shape(x))

        # 2nd branch
        # prepare the Newton's method
        def f(t):
            return t - 1 + alpha * q * t ** (q - 1)

        def df(t):
            return 1 + alpha * q * (q - 1) * t ** (q - 2)

        # use the Newton's method
        root = np.abs(x[mask]) / 2
        low = 5 * 1e-10
        root = newton_(f, df, root, low=low, high=np.inf)

        # explicit solution in the case q==0.5
        mask2 = (q == 0.5) * (alpha <= (2 * (1 - q) / (2 - q)) ** (1 - q) / (2 - q))
        s = (
            2
            * np.sin((np.arccos(3 * np.sqrt(3) * alpha[mask2] / 4) + np.pi / 2) / 3)
            / np.sqrt(3)
        )
        root[mask2] = s**2

        # compute the prox
        prox_x[mask] = root * x[mask]
        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        return np.sum(np.abs(x) ** self.q)

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its components if it"
                + " is an array) must be strictly positive"
            )
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise ValueError("'gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 (06-10-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class PRelu:
    r"""Computes the proximity operator and the evaluation of f.

    Where f is the 'PRelu activation' function defined as:

                          /   0                            if  x> 0
                   f(x)= |
                         \   (1/alpha +1) * x^2 /2         otherwise


    When the input 'x' is an array, the output is computed element-wise :

    -When calling the function, the output is a scalar (sum of the element-wise
    results ).

    - But for the proximity operator (method 'prox'), the output has the
    same shape as the input 'x'.


    Note: The proximity operator of the function defined below is the PRelu function

     INPUTS
    ========
     x       - scalar or ND array
     alpha   - positive, scalar or ND array with the same size as 'x'  in ]0, 1]
                [default value: alpha=0.25]

    =======
    Examples
    ========

     Evaluate the function  f:

     >>> PRelu()( -2 )
      10.0

      Compute the result element-wise sum for vector inputs :

     >>> PRelu(alpha=0.5)( [-1, -3, 2] )
     15.0

     Compute the proximity operator at a given point :

     >>> PRelu().prox(  [-2, 3, 4 ])
      array([-0.5,  3. ,  4. ])
     >>> PRelu( [0.5, 0.25, 0.3]).prox( [-2, 3, 4 ] )
     array([-1.,  3.,  4.])
    """

    def __init__(self, alpha: Union[float, np.ndarray] = 0.25):
        if np.size(alpha) > 1 and (not isinstance(alpha, np.ndarray)):
            alpha = np.array(alpha)
        if np.any(alpha <= 0) or np.any(alpha > 1):
            raise ValueError(
                "'alpha (or all its components if it is an array) must be in ]0, 1]"
            )
        self.alpha = alpha

    def prox(self, x: np.ndarray) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        if np.isscalar(x):
            if x > 0:
                return x
            else:
                return self.alpha * x
        prox_x = self.alpha * x
        mask = x > 0
        prox_x[mask] = prox_x[mask]
        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        fun_x = np.zeros(np.shape(x))
        mask = x <= 0
        alpha_ = self.alpha
        if np.size(self.alpha) > 1:
            alpha_ = self.alpha[mask]

        fun_x[mask] = (1 / alpha_ + 1) * x[mask] ** 2 / 2

        return np.sum(fun_x)



"""
Version : 1.0 ( 06-14-2022).

DEPENDENCIES:
     -  'L1Ball.py' - located in the folder 'indicator'
     - 'Linf.py'    - located in the folder 'multi'

Author  : Mbaye DIONGUE

Copyright (C) 2019

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Optional
import numpy as np
from proxop.multi.Linf import Linf


class L1Ball:
    """Compute the projection and the indicator of the L1 ball.

    Recall: every vector X belonging to the L1 ball verifies:

                        ||x||_1 <= eta

     INPUTS
    ========
     x         - ND array
     eta       - positive scalar
     axis - int or None, direction of block-wise processing
            When the input 'x' is an array, the computation can vary as follows:
            - axis = None --> 'x' is processed as a single vector [DEFAULT]
            - axis >= 0 --> 'x' is processed block-wise along the specified axis
               (axis=0 -> rows, axis=1 -> columns etc.).
    """

    def __init__(
            self,
            eta: float,
            axis: Optional[int] = None
    ):
        if np.size(eta) > 1 or eta <= 0:
            raise Exception("'eta' must be a positive scalar")
        self.eta = eta
        self.gamma = 1
        self.axis = axis

    def prox(self, x: np.ndarray) -> np.ndarray:
        return x - Linf(axis=self.axis).prox(x, gamma=self.eta)

    def __call__(self, x: np.ndarray) -> float:
        """
        Indicate if the input 'x' is in the constraint set or not.

        Parameters
        ----------
        x : np.ndarray

        Returns
        -------
        0      if the input 'x' is in the set
        +inf   otherwise
        """

        tol = 1e-10
        l1_x = np.linalg.norm(x, ord=1, axis=self.axis)
        if np.all(l1_x <= self.eta + tol):
            return 0
        return np.inf



"""
Version : 1.0 (06-09-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

import numpy as np


class TanhActi:
    r"""Computes the proximity operator and the evaluation of f.

    Where f is the 'Hyperbolic tangent activation' function defined as:

                  / x*arctan(x)+(log(1-x^2)-x^2)/2   if |x| < 1
          f(x) = |
                 \ +Inf                              otherwise


    When the input 'x' is an array, the output is computed element-wise :

    -When calling the function, the output is a scalar (sum of the element-wise
    results).

    - But for the proximity operator (method 'prox'), the output has the same
    shape as the input 'x'.

     INPUTS
    ========
     x     - scalar or ND array

    ========
    Examples
    ========

     Evaluate the function  f:

     >>> TanhActi()( 0.5 )
     0.00581203594

      Compute the result element-wise for vector inputs :

     >>> TanhActi()([ 0.3, 0.25, 0.5] )
     0.0068465198684131146

     Compute the proximity operator at a given point :

     >>> TanhActi().prox(  [-2, .3, 4 ])
     array([-0.96402758,  0.29131261,  0.9993293 ])
    """

    def __init__(self):
        pass

    def prox(self, x: np.ndarray) -> np.ndarray:
        return np.tanh(x)

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.isscalar(x):
            x = np.reshape(x, (-1))
        prox_x = np.zeros(np.shape(x))
        mask = np.abs(x) < 1
        prox_x[mask] = x[mask] * np.arctanh(x[mask]) + 0.5 * (
            np.log(1 - x[mask] ** 2) - x[mask] ** 2
        )
        prox_x[np.abs(x) >= 1] = np.inf
        return np.sum(prox_x)



"""
Version : 1.0 ( 06-21-2022).

DEPENDENCIES:
     - 'prox_svd.py' located in the folder 'utils'
     - 'fun_svd.py' located in the folder 'utils'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from proxop.utils.prox_svd import prox_svd
from proxop.utils.fun_svd import fun_svd
import numpy as np


class PermutationInvariant:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the function defined as:

                            f(X) = phi(s)

    where :
          * X = U*diag(s)*V.T \in R^{M*N}  (the Singular Value decomposition of X)

          * 'gamma' is the scale factor

     INPUTS
    ========
     x           -  (M,N) -array_like ( representing an M*N matrix)
     gamma       - positive scalar  [default: gamma=1]
     ClassPhi    - A class that has at least the two methods: 'ClassPhi.__call___(x)'
                   and 'phi.prox(x, tau)'. >>> ClassPhi.__call__(x) must return phi(x)
                   (i.e. the value of the function phi evaluated at  x) and
                   '>>> ClassPhi.prox(x, tau)' the proximity operator of tau*phi
                   evaluated at x.
    """

    def __init__(
            self,
            class_phi
    ):
        self.class_phi = class_phi

    def prox(self, x: np.ndarray, gamma: float = 1) -> np.ndarray:
        if np.any(gamma <= 0) or np.size(gamma) > 1:
            raise Exception("'gamma'  must be a strictly positive scalar")

        return prox_svd(x, gamma, self.class_phi().prox)

    def __call__(self, x) -> float:
        return fun_svd(x, 1, self.class_phi().prox)



"""
Original Version : 1.0 updated  (06-09-2022).

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

import numpy as np


class ArctanActi:
    r"""Compute the proximity operator and the evaluation of f.

     where f is the arc tangent function defined as:

                 / -(2/pi)*log(cos(pi*x/2))-(x^2)/2   if |x| < 1
         f(x) = |
                \ +Inf                 otherwise

    When the input 'x' is an array, the output is computed element-wise :

    - When calling the function (not the proximity operator) the result
    is computed element-wise SUM. So the command >>>ArctanActi()(x) will
    return a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the
    same shape as the input 'x'. So, the command >>>ArctanActi().prox(x)
    will return an array with the same shape as 'x'


     INPUTS
    ========
     x     - scalar or ND array

    ========
    Examples
    ========

    Evaluate the 'direct' function :
    >>> ArctanActi()(-0.5)
    0.09563560015265157
    >>>ArctanActi()([-0.2, -0.5, 0, 0.25, 0.3])
    0.155203962389311

    Compute the proximity operator at a given point:
    >>> ArctanActi().prox( 3)
    0.7951672353008666
    >>> ArctanActi().prox([ -3., 0.5, 6.])
    array([-0.79516724,  0.29516724,  0.89486309])
    """

    def __init__(self):
        pass

    def prox(self, x: np.ndarray) -> np.ndarray:
        return 2 / np.pi * np.arctan(x)

    def __call__(self, x: np.ndarray) -> float:
        if not isinstance(x, np.ndarray):
            x = np.array(x)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))

        fun_x = np.zeros(np.size(x))
        mask = abs(x) < 1
        fun_x[mask] = -(
            (2 / np.pi) * np.log(np.cos(np.pi * x[mask] / 2)) - x[mask] ** 2 / 2
        )
        fun_x[np.abs(x) >= 1] = np.inf
        return np.sum(fun_x)



"""
Version : 1.0 ( 06-10-2022).

DEPENDENCIES:
     - 'newton.py' located in the folder 'utils'

Author  : Mbaye DIONGUE

Copyright (C) 2019

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Optional, Tuple
import numpy as np
from proxop.utils.newton import newton_


class EpiDistance:
    r"""Compute the projection and the indicator of the epigraph of phi.

    Where phi is the function defined as:

            phi(y) = w * ||y - P_C(y)||_2^q       with q=1 or q=2


     The caller must ensure that 'project_C' uses the same type of processing.

      INPUTS
     ========
      y      - ND array
      ksi     - ND array compatible with the blocks of 'y'
      w       - positive, scalar or ND array with the same size as 'ksi'
      axis    - int or None, axis of block-wise processing [OPTIONAL]
               When the input 'y' is an array, the computation can vary as follows:
              - axis = None --> 'y' is processed as a single vector [DEFAULT]
                            (in this case, 'xi' must be scalar)
              - axis >= 0 --> 'y' is processed block-wise along the specified axis
                            (in this case, 'xi' must be singleton along 'axis')
              - axis < 0 --> 'y' is processed element-by-element.
                            (in this case, 'xi' must be the same size as 'y')
      q         - integer, scalar [OPTIONAL, default: 2]
      project_C - function handle with an argument at least
                [OPTIONAL, default: project_C(x)=0 for all x]
    """

    def __init__(
            self,
            ksi: np.ndarray,
            axis: Optional[int] = None,
            project_C=lambda y: 0,
            q: int = 2,
            w: Union[float, np.ndarray] = 1.0
    ):

        if np.any(w <= 0):
            raise ValueError("'w' must be positive")
        if np.size(w) > 1 and (np.size(w) != np.size(ksi)):
            raise ValueError(" 'w' must be a scalar or must" +
                             " have the same size as 'ksi'")

        if np.size(ksi) <= 1:
            ksi = np.reshape(ksi, (-1))

        self.w = w
        self.ksi = ksi
        self.axis = axis
        self.q = q
        self.project_C = project_C

    def prox(self, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Compute the proximal operator (i.e. the projection on the constraint set).
        """
        self._check(y)
        ksi = self.ksi
        w = self.w
        axis = self.axis
        project_C = self.project_C
        q = self.q
        eps = 1e-16
        sz = np.shape(y)

        if np.size(y) <= 1:
            y = np.array([y])
        if self.axis is None:
            y = np.reshape(y, (-1))

        # preliminaries
        py = project_C(y)
        dy = np.linalg.norm(y - py, ord=2, axis=axis)  # distance separating y from C

        if q == 1:
            t = np.maximum(0, dy + w * ksi) / (1 + w**2)
        else:
            # Finding the root of with the Newton's method
            def fun_phi(u):
                return (
                        q * w ** 2 * u ** (2 * q - 1)
                        - q * w * ksi * u ** (q - 1)
                        + u
                        - np.abs(y)
                )

            def derive_phi(u):
                return (
                        (2 * q - 1) * q * w ** 2 * u ** (2 * q - 2)
                        - (q - 1) * q * w * ksi * u ** (q - 2)
                        + 1
                )

            # starting point
            low = (np.maximum(ksi, 0) / w) ** (1 / q)
            t = newton_(fun_phi, fp=derive_phi, x0=3 * low, low=low, high=np.inf)

        # 3rd branch
        prox_p = py + t / (dy + eps) * (y - py)

        prox_t = w * np.linalg.norm(prox_p - project_C(prox_p), ord=2, axis=axis) ** q

        # default cas: axis==None => y is processed as single vector
        if axis is None:
            if dy == 0 and ksi < 0:
                return tuple([y, np.zeros_like(ksi)])
            if w * dy**q <= ksi:
                return tuple([y, ksi])
            return tuple([prox_p, prox_t])

        # 2nd branch
        mask_p = np.ones(np.shape(y), dtype=bool)
        mask_t = np.logical_and(dy == 0, ksi < 0)
        mask_p = mask_t * mask_p
        prox_p[mask_p] = y[mask_p]
        prox_t[mask_t] = 0

        # 1st branch

        mask_t = w * dy**q <= ksi
        mask_p = mask_t * mask_p
        prox_p[mask_p] = y[mask_p]
        prox_t[mask_t] = ksi[mask_t]

        prox_p = np.reshape(prox_p, sz)
        return tuple([prox_p, prox_t])

    # indicator of the constraint set
    def __call__(self, y: np.ndarray) -> float:
        self._check(y)
        TOL = 1e-10
        if self.axis is None:
            y = np.reshape(y, (-1))
        py = self.project_C(y)
        dy = np.linalg.norm(y - py, ord=2, axis=self.axis)

        if np.all(self.w * dy**self.q <= self.ksi + TOL):
            return 0
        return np.inf

    def _check(self, y: np.ndarray):
        sz_ = 1
        if self.axis is not None:
            sz = np.array(np.shape(y), dtype=int)
            sz_ = np.delete(sz, self.axis)

        cond1 = self.axis is not None and np.any(sz_ != np.shape(self.ksi))
        cond2 = self.axis is None and np.any(np.size(self.ksi) != sz_)
        if cond1 or cond2:
            try:
                self.ksi = np.reshape(self.ksi, sz_)
            except ValueError:
                print("'ksi' is not compatible with the size of 'y' ")



"""
Version : 1.0 (06-09-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np
from proxop.utils.newton import newton_


class NegativeRoot:
    r"""Computes the proximity operator and the evaluation of gamma*f.

    Where f is the function defined as:


                  /  - x^(1/q)    if x>=0
           f(x)= |
                 \    +INF                 otherwise

                 with q>=1

    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise :

    -When calling the function, the output is a scalar (sum of the element-wise
    results ).

    - But for the proximity operator (method 'prox'), the output has the same
    shape as the input 'x'.

     INPUTS
    ========
     x     - scalar or ND array
     q     - scalar >=1
     gamma - positive, scalar or ND array with the same size as 'x' (default: gamma=1)

    =======
    Examples
    ========

     Evaluate the function  f:

     >>> NegativeRoot(3)( 8 )
     -2.0

      Compute the result element-wise sum for vector input :

     >>> NegativeRoot(q=2, gamma=[1, 2, 3])( [1, 16, 9] )
     -18.0

     Compute the proximity operator at a given point :

     >>> NegativeRoot(2).prox(  [-2, 3, 4 ])
     array([0.49278772, 1.3453765 , 1.43519754])

     Use a scale factor 'gamma'>0 to compute the proximity operator of  the
     function 'gamma*f'

     >>> NegativeRoot( 2, gamma=[1,2,3,3]).prox( [-2, 3, 4, np.e ] )
     array([0.49278772, 1.37090672, 1.47180413, 1.36953862])
    """

    def __init__(self, q: float):
        if np.any(q < 1) or np.size(q) > 1:
            raise ValueError("'q' must be a scalar greater than 1")
        self.q = q

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)

        self._check(x, gamma)
        scale = gamma

        abs_x = np.abs(x)
        q = self.q
        if q == 1:
            return np.maximum(0, x + scale)

        def polynom_phi(t):
            return t ** (2 * q - 1) - x * t ** (q - 1) - scale / q

        def der_phi(t):
            return (2 * q - 1) * t ** (2 * q - 2) - (q - 1) * x * t ** (q - 2)

        # starting point
        prox_x = abs_x

        # Finding the root of the polynom with Newton method
        prox_x = newton_(polynom_phi, fp=der_phi, x0=prox_x, low=0, high=np.inf)

        return prox_x ** (1 / q)

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        fun_x = np.zeros(np.shape(x))
        mask = x >= 0
        fun_x[mask] = -x[mask] ** (1 / self.q)
        fun_x[np.logical_not(mask)] = np.inf

        return np.sum(fun_x)

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its components if it is an array) must "
                + "be strictly positive"
            )
        if (np.size(gamma) > 1) and (np.shape(gamma) != np.shape(x)):
            raise ValueError("gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 ( 06-15-2022).

DEPENDENCIES:
     - 'newton.py' located in the folder 'utils'

Author  : Mbaye DIONGUE

Copyright (C) 2019

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Tuple
import numpy as np
from proxop.utils.newton import newton_


class EpiPower:
    r"""Compute the projection and the indicator of the epigraph of phi.

    Where phi is the function defined as:

                phi(y)= w * |y|^ q

    When the inputs are arrays, the outputs are computed element-wise
     INPUTS
    ========
     y    - ND array
     ksi  - ND array with the same size as 'y'
     w    - positive, scalar or ND array with the same size as ksi [default: w=1.0]
     q    - positive scalar [default: q=2]
    """

    def __init__(
            self,
            ksi: np.ndarray,
            q: float = 2,
            w: Union[float, np.ndarray] = 1.0
    ):

        if np.any(w <= 0):
            raise ValueError("'w' must be positive")
        if np.size(q) > 1 or q <= 0:
            raise ValueError("'q' (or must be a positive scalar")
        if np.size(w) > 1 and (np.size(w) != np.size(ksi)):
            raise ValueError(" 'w' must be a scalar or have the same size as 'ksi'")
        if np.size(ksi) <= 1:
            ksi = np.reshape(ksi, (-1))

        self.q = q
        self.w = w
        self.ksi = ksi

    # proximal operator (i.e. the projection on the constraint set)
    def prox(self, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        self._check(y)
        ksi = self.ksi
        w = self.w
        q = self.q
        if np.size(y) <= 1:
            y = np.reshape(y, (-1))

        # 2nd branch
        def polynom_phi(t):
            return (
                q * w**2 * t ** (2 * q - 1)
                - q * w * ksi * t ** (q - 1)
                + t
                - np.abs(y)
            )

        def der_phi(t):
            return (
                (2 * q - 1) * q * w**2 * t ** (2 * q - 2)
                - (q - 1) * q * w * ksi * t ** (q - 2)
                + 1
            )

        # starting point
        low = (np.maximum(ksi, 0) / w) ** (1 / q)
        root = 3 * low

        # Finding the root of the polynom with the Newton method
        root = newton_(polynom_phi, fp=der_phi, x0=root, low=low, high=np.inf)
        prox_t = w * root ** q
        prox_y = np.sign(y) * root

        # 1st branch
        mask = w * np.abs(y) ** q <= ksi
        prox_y[mask] = y[mask]
        prox_t[mask] = ksi[mask]
        return tuple([prox_y, prox_t])

    # indicator of the constraint set
    def __call__(self, y: np.ndarray) -> float:
        tol = 1e-10
        if np.all(self.w * np.abs(y) <= self.ksi) + tol:
            return 0
        return np.inf

    def _check(self, y):
        if np.size(y) != np.size(self.ksi):
            raise ValueError(" 'y' must have the same size as 'ksi'")



"""
Version : 1.0 (06-09-2022).

DEPENDENCIES:
    - 'Thresholder.py' located in the folder 'scalar'

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from proxop.scalar.Thresholder import Thresholder


class AbsValue(Thresholder):
    """Compute the proximity operator and the evaluation of the gamma*f.

      where the function f is defined as:

                        f(x) = |x|

     'gamma' is the scale factor

     When the input 'x' is an array, the output is computed element-wise :

     -When calling the function (and not the proximity operator) the result
     is computed element-wise SUM. So the command >>>AbsValue()(x) will
     return a scalar even if x is a vector.

     -But for the proximity operator (method 'prox'), the output has the same
     shape as the input 'x'.
     So, the command >>>AbsValue().prox(x)   will return an array with the same
     shape as 'x'


      INPUTS
     ========
     x     - scalar or ND array
     gamma - positive, scalar or ND array with the same size as 'x'
     [default: gamma=1]


     ========
     Examples
     ========

     >>> AbsValue()(-3)
     3

     >>> AbsValue().prox( 3)
     2
     >>> AbsValue().prox([ -3., 1., 6.])
     array([-2.,  0.,  5.])


    Use a scale factor 'gamma'>0 to compute the proximity operator of  the
    function 'gamma*f'

     >>> AbsValue().prox([ -3., 1., 6.], gamma=2)
      array([-1.,  0.,  4.])
    """

    def __init__(self):
        super().__init__(-1, 1)



"""
Version : 1.0 ( 06-21-2022).

DEPENDENCIES:
     - 'prox_svd.py' located in the folder 'utils'
     - 'prox_svd.py' located in the folder 'utils'
     - 'newton.py'  in the folder 'utils'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

import numpy as np
from proxop.utils.prox_svd import prox_svd
from proxop.utils.fun_svd import fun_svd
from proxop.utils.newton import newton_


class InvPenaltyLogDet:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the sum of the Inverse Schatten Penalty and logDeterminant function
    defined as:

               / -log( det(X) ) +  R_p( X^{-1} )p   if X is a symmetric positive
         f(x)=|                                     definite  matrix
              \  + inf                              otherwise

    where

           * X = U*diag(s)*V.T \in R^{M*N}  the Singular Value decomposition of X

           * R_p(X^{-1})= ||s||_p^p  is the p-Schatten norm of the inverse of X
           (the inverse X exists because X is a symmetric definite positive matrix)

           * det(X)= prod_i(s_i)    is the determinant of the matrix X

           * 'gamma' is the scale factor

    Note:No checking is performed to verify whether X is symmetric or not when computing
    the proximity operator with the method 'prox'. X is assumed to be symmetric.

     INPUTS
    ========
     x               -  (M,N) -array_like ( representing an M*N matrix)
     p               -  scalar in ]0, 1]
     mu              - positive scalar [default: mu=1]
     gamma           - positive scalar  [default: gamma=1]
    """

    def __init__(
            self,
            p: float,
            mu: float = 1
    ):
        if np.any(mu < 0) or np.size(mu) > 1:
            raise Exception("'mu'  must be a positive scalar")
        if np.any(p < 0) or np.any(p > 1) or np.size(p) > 1:
            raise Exception("'p'  must be a scalar belonging to ]0, 1] ")
        self.mu = mu
        self.p = p

    def prox(self, x: np.ndarray, gamma: float = 1) -> np.ndarray:
        self._check(x, gamma)
        scale = gamma
        mu = self.mu
        p = self.p

        def prox_phi(s, gam):
            # Note: same computation as in 'EpiPower.py' for q=2
            def polynom_phi(t):
                return t ** (p + 2) - s * t ** (p + 1) - gam * t**p - mu * gam * p

            def derive_polynom_phi(t):
                return (
                    (p + 2) * t ** (p + 1)
                    - (p + 1) * s * t**p
                    - p * gam * t ** (p - 1)
                )

            # starting point
            low = 1e-10
            root_init = np.abs(s)
            # Finding the root of the polynom with the Newton method
            d = newton_(
                polynom_phi, fp=derive_polynom_phi, x0=root_init, low=low, high=np.inf
            )
            return d

        return prox_svd(x, scale, prox_phi, hermitian=True)

    def __call__(self, x: np.ndarray) -> float:
        self._check(x)
        p = self.p
        # Check if the matrix is symmetric
        if not np.allclose(x, np.transpose(x)):
            return np.inf

        def fun_phi(s):
            if np.any(s == 0):
                return np.inf
            return -np.log(np.prod(s)) + self.mu * np.sum((np.abs(1 / s)) ** p)

        return fun_svd(x, 1, fun_phi)

    def _check(self, x, gamma=1):
        if np.any(gamma <= 0) or np.size(gamma) > 1:
            raise Exception("'gamma'  must be a strictly positive scalar")
        if np.size(gamma) > 1 and np.size(gamma) != np.size(x):
            ValueError(
                "'gamma' must be a positive scalar or a positive ND array" +
                " with the same size as 'x'"
            )



"""
Version : 1.0 ( 06-13-2022).

DEPENDENCIES:
     - 'solver_cubic.py' located in the folder 'utils'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Optional, Tuple
import numpy as np
from proxop.utils.solver_cubic import solver_cubic


class PersAbsDiff:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the function defined as:


                 /   ||y||_2^2 / xi          if xi > 0
      f(y,xi) = |  0                        if y = 0 and xi = 0
                \  +inf                     otherwise


    'gamma' is the scale factor

    INPUTS
    ========
    y         - scalar or ND array
    xi        - scalar or ND array compatible with the blocks of 'y'
    gamma     - positive, scalar or ND array compatible with the blocks of 'x'
                 [default: gamma=1]
    axis      - None or int, axis of block-wise processing [default: axis=None]
                  axis = None --> 'x' is processed as a single vector [DEFAULT] In this
                  case 'gamma' must be a scalar.
                  axis >=0   --> 'x' is processed block-wise along the specified axis
                  (0 -> rows, 1-> columns ect. In this case 'gamma' and 'xi' must be
                  singletons along 'axis'.
    """

    def __init__(
            self,
            xi: Union[float, np.ndarray] = 1,
            axis: Optional[int] = None
    ):
        if (axis is not None) and (axis < 0):
            axis = None
        if np.size(xi) <= 1:
            xi = np.reshape(xi, (-1))
        self.axis = axis
        self.xi = xi

    def prox(
            self,
            y: np.ndarray,
            gamma: Union[float, np.ndarray] = 1
    ) -> Tuple[np.ndarray, np.ndarray]:

        self._check(y, gamma)
        if np.size(gamma) <= 1:
            gamma = np.reshape(gamma, (-1))
        scale = gamma
        axis = self.axis
        xi = self.xi
        sz0 = np.shape(y)
        if np.size(y) <= 1:
            y = np.reshape(y, (-1))

        sz = np.shape(y)
        sz = np.array(sz, dtype=int)
        sz[axis] = 1

        if np.size(scale) > 1:
            scale = np.reshape(scale, sz)
        if np.size(xi) > 1:
            xi = np.reshape(xi, sz)

        # linearize
        if axis is None:
            y = np.reshape(y, (-1))

        # 3rd branch
        l2_x = np.sqrt(np.sum(y**2, axis=axis)).reshape(sz)
        t = solver_cubic(scale, 0, 4 * xi + 8 * scale, -8 * l2_x)[0]
        # t= np.max( np.real(t), axis=axis).reshape(sz)
        pp = scale * t / l2_x
        qq = scale * t**2 / 4

        # 1st branch
        mask = l2_x**2 <= -4 * scale * xi
        pp[mask] = 1
        qq[mask] = -xi[mask]

        # 2nd branch
        mask = np.logical_and(l2_x == 0, xi > 0).reshape(sz)
        pp[mask] = 1
        qq[mask] = 0

        # compute the prox
        prox_p = y - y * pp
        prox_q = xi + qq

        # revert back
        prox_p = np.reshape(prox_p, sz0)
        prox_q = np.reshape(prox_q, np.shape(self.xi))

        return tuple([prox_p, prox_q])

    def __call__(self, y: np.ndarray) -> float:
        if np.size(y) <= 1:
            y = np.reshape(y, (-1))

        if np.any(self.xi < 0):
            return np.inf
        l2_y2 = np.sum(y**2, axis=self.axis)
        fun_y = l2_y2 / self.xi
        fun_y[self.xi == 0] = 0

        return np.sum(fun_y)

    def _check(self, x, gamma):
        sz = np.shape(x)
        if np.any(gamma <= 0):
            raise Exception("'gamma'  must be strictly positive")

        if self.axis is None and np.size(gamma) > 1:
            raise Exception(
                "A 'axis' must be specified when 'gamma' is not a scalar"
            )
        if len(sz) <= 1:
            self.axis = None
        if len(sz) <= 1 and (np.size(gamma) > 1 or np.size(self.xi) > 1):
            raise Exception(
                "'gamma' and 'xi' must be scalar when 'x' is one dimensional"
            )

        if len(sz) > 1 and self.axis is not None:
            sz = np.array(sz, dtype=int)
            sz[self.axis] = 1
            if 1 < np.size(gamma) != np.prod(sz):
                raise Exception(
                    "The dimension of 'gamma' is not compatible with the blocks of 'x'"
                )
            if np.size(self.xi) > 1 and (np.prod(sz) != np.size(self.xi)):
                raise Exception(
                    "The dimension of 'xi is not compatible with the blocks of 'x'"
                )



"""
The function solve a 3rd degree polynomial equation.

    a * x^3 + b * x^2 + c * x + d = 0

Evaluate phi(s) where s is the vector that is composed by singular values of x.

Version : 1.0 (06-14-2022).

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


def solver_cubic(
        a: Union[float, np.ndarray],
        b: Union[float, np.ndarray],
        c: Union[float, np.ndarray],
        d: Union[float, np.ndarray]
) -> Union[float, np.ndarray]:

    if np.any(a == 0):
        raise Exception("A cubic equation is expected: a cannot be null")
    if np.isscalar(a):
        a = np.array([a])
        b = np.array([b])
        c = np.array([c])
        d = np.array([d])

    # reduction to a depressed cubic (t^3 + p*t + q = 0)
    p = c / a - b**2 / (3 * a**2)
    q = (2 * b**3 / 27 - a * b * c / 3 + a**2 * d) / a**3

    # discriminant
    dd = (p / 3) ** 3 + (q / 2) ** 2

    # output vectors
    y1 = np.zeros_like(dd)  # the first root is always real
    y2 = np.zeros_like(dd, dtype=complex)
    y3 = np.zeros_like(dd, dtype=complex)

    # 1st case: 3 real unequal roots
    idx = dd < 0
    phi = np.arccos(-q[idx] / 2 / np.sqrt(np.abs(p[idx] ** 3 / 27)))
    tau = 2 * np.sqrt(np.abs(p[idx] / 3))
    y1[idx] = tau * np.cos(phi / 3)
    y2[idx] = -tau * np.cos((phi + np.pi) / 3)
    y3[idx] = -tau * np.cos((phi - np.pi) / 3)

    # 2nd case: 1 real root + 2 conjugate complex roots
    idx = dd > 0
    z1 = -q[idx] / 2
    z2 = np.sqrt(dd[idx])
    u = np.cbrt(z1 + z2)
    v = np.cbrt(z1 - z2)
    y1[idx] = u + v
    e1 = (-1 + 1j * np.sqrt(3)) / 2
    e2 = (-1 - 1j * np.sqrt(3)) / 2
    y2[idx] = u * e1 + v * e2
    y3[idx] = u * e2 + v * e1

    # 3rd case: 1 simple real root + 1 double real root
    idx = dd == 0
    y1[idx] = 3.0 * q[idx] / p[idx]
    y2[idx] = -1.5 * q[idx] / p[idx]
    y3[idx] = y2[idx]

    # correction to the original cubic
    t = b / (3 * a)
    y1 = y1 - t
    y2 = y2 - t
    y3 = y3 - t

    return [y1, y2, y3]



"""
Version : 1.0 ( 06-23-2022).

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class BregBoltzShannon:
    r"""Compute the proximity operator prox_{gamma*f}^{phi}.

    Where the functions f and phi are defined as follows:

                        /  x*log(x) - w*x      if x>0
               f(x) =  |
                       \  +inf                 otherwise


                         /  u*log(u)-u     if u>0
              phi(u) =  |
                        \   +inf           otherwise

    Note: The function phi is a Legendre type function and induces a Bregman distance

    'gamma' is the scale factor

     When the input 'x' is an array, the output is computed element-wise :

    - When calling the functions f or phi (method 'phi') the result is computed
    element-wise SUM. So the command >>>BregBoltzShannon()(x) or
    >>BregBoltzShannon().phi(x) will give a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the same
    shape as the input 'x'. So, the command >>>BregBoltzShannon.prox(x)
    will return an array with the same shape as 'x'


    INPUTS
    ========
    x      - scalar or ND array
    w      - scalar or ND array with the same size as 'x'  [default: w=1]
    gamma  - positive, scalar or ND array with the same size as 'x' [default: gamma=1]

     ========
     Examples
     ========

     Evaluate the  function f :

     >>> BregBoltzShannon()(1)
     -1.0
     >>> BregBoltzShannon(w=0)( [0.2, 3, 4, np.e] )
     11.237408556456117

     Evaluate The function  phi

     >> BregBoltzShannon(w=0).phi( [0.2, 3, 4, np.e] )
     1.3191267279970713

     Compute the proximity operator at a given point :

     >>> BregBoltzShannon(2).prox( [-2, 3, 4] )
     array([  0.36787944,  54.59815003, 148.4131591 ])
     >>> BregBoltzShannon(w=1).prox( [-2, 3, 4], gamma=2 )
     array([ 0.13533528, 20.08553692, 54.59815003])
    """

    def __init__(self, w: Union[float, np.ndarray] = 1.0):
        if np.size(w) > 1 and (not isinstance(w, np.ndarray)):
            w = np.array(w)
        self.w = w

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        self._check(x, gamma)
        w = self.w
        return np.exp((x + gamma * (w - 1)))

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        self._check(x, 1)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        res = np.zeros(np.shape(x))
        mask = x > 0
        w = self.w
        if np.size(w) > 1:
            w = w[mask]
        res[mask] = x[mask] * np.log(x[mask]) - w * x[mask]
        res[x <= 0] = np.inf
        return np.sum(res)

    def phi(self, u: np.ndarray) -> float:
        if np.size(u) > 1 and (not isinstance(u, np.ndarray)):
            u = np.array(u)
        if np.size(u) <= 1:
            u = np.reshape(u, (-1))
        res = np.zeros(np.shape(u))
        mask = u > 0
        res[mask] = u[mask] * np.log(u[mask]) - u[mask]
        res[u <= 0] = np.inf
        return np.sum(res)

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError("'gamma'  must be strictly positive")
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise ValueError("gamma' must be either scalar or the same size as 'x'")
        if (np.size(self.w) > 1) and (np.size(self.w) != np.size(x)):
            raise ValueError("'w' must be either scalar or the same size as 'x'")



"""
Version : 1.0 (06-09-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class FermiDiracEntropy:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the Fermi Dirac Entropy function defined as:

                      /  ( x*log(x)  + (1-x)log(x) )    if  0 < x < 1
               f(x)= |  0                               if x in {0, 1}
                     \  +INF                            otherwise


    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise :

    -When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>FermiDIracEntropy(x) will
    return a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the same
    shape as the input 'x'. So, the command >>>FermiDIracEntropy.prox(x)
    will return an array with the same shape as 'x'.


    INPUTS
    ========
     x     - scalar or array
     gamma  - positive, scalar or array with the same size as 'x' [default: gamma=1]

     =======
    Examples
    ========

     Evaluate the 'direct' function :

     >>> FermiDiracEntropy()( 0.5 )
     -0.6931471805599453

      Compute the result element-wise sum for vector inputs :

     >>> FermiDiracEntropy(gamma=2)( [1, 0, 0.25, 0.5] )
     -2.510964650357507

     Compute the proximity operator at a given point :

     >>> FermiDiracEntropy().prox(  [-2, 3, 4, np.e] )
     array([0.10829336, 0.89170664, 0.9545842 , 0.86456313])

     Use a scale factor 'gamma'>0 to compute the proximity operator of  the function
     'gamma*f' :

     >>> FermiDiracEntropy().prox( [-2, 3, 4, np.e ], gamma=2.5)
     array([0.28609273, 0.71390727, 0.78356663, 0.69220157])
    """

    def __init__(self):
        pass

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        self._check(x, gamma)

        limit = 8
        # RUN HALEY'S METHOD FOR SOLVING w = W_{exp(x/gamma)}(exp(x/gamma)/gamma)
        # where W_r(x) is the generalized Lambert function

        w = np.zeros_like(x)
        igamma = 1 / gamma
        c = x * igamma - np.log(gamma)
        z = np.exp(c)
        r = np.exp(x * igamma)

        # ASYMPTOTIC APPROX
        approx = 1 - np.exp((1 - x) * igamma)

        # INITIALIZATION
        # Case 1: gamma <= 1/30
        mask = np.logical_and(z > 1, gamma <= 1 / 30)
        w[mask] = c[mask] - np.log(c[mask])
        # Case 2: gamma > 1/30
        mask = np.logical_and(z > 1, gamma > 1 / 30)
        if np.size(igamma) > 1:
            igamma = igamma[mask]
        w[mask] = igamma * approx[mask]

        # RUN
        maxiter = 20
        test_end = np.zeros_like(x)
        tol = 1e-8
        epsilon = 1e-16  # to avoid dividing by zero
        n = np.size(x)
        for _ in range(maxiter):
            e = np.exp(w)
            y = w * e + r * w - z
            v = e * (1 + w) + r
            u = e * (2 + w)
            wnew = w - y / (v - y * u / (2 * v))

            mask = np.abs(wnew - w) / (np.abs(w) + epsilon) < tol
            test_end[np.logical_and(mask, test_end == 0)] = 1
            not_mask = np.logical_not(mask)
            idx_update = np.logical_and(not_mask, test_end == 0)
            w[idx_update] = wnew[idx_update]  # the rest stays constant !
            if np.sum(test_end) == n:  # stop !
                break

        prox_x = gamma * w
        if np.size(prox_x) <= 1:
            if c > limit and gamma > 1:
                prox_x = approx
            return np.minimum(prox_x, 1)

        # ASYMPTOTIC DVP
        mask_approx = np.logical_and(c > limit, gamma > 1)
        prox_x[mask_approx] = approx[mask_approx]

        # FINAL TRESHOLD TO AVOID NUMERICAL ISSUES FOR SMALL GAMMA
        prox_x = np.minimum(prox_x, 1)
        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        result = np.zeros(np.shape(x))
        mask = np.logical_and(x > 0, x < 1)
        result[mask] = x[mask] * np.log(x[mask]) + (1 - x[mask]) * np.log(1 - x[mask])
        mask2 = np.logical_or(x < 0, x > 1)
        result[mask2] = np.inf
        return np.sum(result)

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its elements if it is an array)"
                + "must be strictly positive"
            )
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise ValueError("gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 ( 06-15-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2019

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

import numpy as np


class LogProjection:
    """Compute the projection and the indicator of the set S.

    Where every vector X belonging to S verifies:

         sum_{n=1}^N  alpha * x_n - z_n + z_n * log(z_n/(alpha*x_n)) <= eta


     INPUTS
    ========
     x         - positive ND-array
     eta       - scalar
     z         - positive ND-array with the same size as 'x'
     alpha     - positive, scalar [OPTIONAL, default: alpha=1]
    """

    def __init__(
            self,
            eta: float,
            z: np.ndarray,
            alpha: float = 1
    ):
        if np.size(eta) > 1:
            raise ValueError("'eta' must be a scalar or an array with one element")
        if np.size(alpha) > 1:
            raise ValueError("'alpha' must be a scalar or an array with one element")
        if np.any(z <= 0):
            raise ValueError("'z' (or all its elements) must be strictly positive")
        self.eta = eta
        self.z = z
        self.alpha = alpha

    def prox(self, x: np.ndarray) -> np.ndarray:
        # target function
        alpha = self.alpha
        z = self.z
        eta = self.eta
        if np.any(x <= 0):
            raise ValueError("'x' (or all its elements) must be strictly positive")
        kld = alpha * x - z + z * np.log(z / (alpha * x))

        # projection
        if np.sum(kld) <= eta:  # x is already in the constraint set: projection(x)=x
            return x

        # convert the upper bound
        lamb = self._get_kld_lambda(x, z, alpha, eta)

        # compute the prox
        a = x - alpha * lamb
        prox_x = 0.5 * (a + np.sqrt(a**2 + 4 * lamb * z))
        return prox_x

    def _get_kld_lambda(self, x, z, alpha, eta):
        # prox of KLD (and its derivative)
        def prox_kld(b):
            return 0.5 * ((x - alpha * b) + np.sqrt((x - alpha * b) ** 2 + 4 * b * z))

        def derive_proxkld(b):
            return (z - prox_kld(b)) / np.sqrt((x - alpha * b) ** 2 + 4 * b * z)

        # discrepancy function (and its derivative)
        c = np.sum(z - z * np.log(z))

        def f(b):
            return np.sum(prox_kld(b) - z * np.log(prox_kld(b))) - c - eta

        def df(b):
            return np.sum(derive_proxkld(b) * (1 - z / prox_kld(b)))

        # find the zero of f
        max_it = 1000
        tol = 1e-7
        lambda_ = 1
        stop = False
        n = 0
        while n < max_it and (not stop):
            lambda_old = lambda_

            # Newton's method
            lambda_ = lambda_ - f(lambda_) / (
                df(lambda_) + tol
            )  # tol is put to avoid dividing by zero

            if np.abs(lambda_ - lambda_old) < tol * np.abs(lambda_old):
                stop = True
            n += 1
        return lambda_

    # Indicator of the constraint set
    def __call__(self, x: np.ndarray) -> float:
        """
        Indicate if the input 'x' is in the constraint set or not.

        Parameters
        ----------
        x : np.ndarray

        Returns
        -------
        0      if the input 'x' is in the set
        +inf   otherwise
        """
        if np.any(x <= 0):
            raise ValueError("'x' (or all its elements) must be strictly positive")
        kld = self.alpha * x - self.z + self.z * np.log(self.z / (self.alpha * x))
        if np.sum(kld) <= self.eta:
            return 0
        return np.inf



"""
Version : 1.0 ( 06-22-2022).

DEPENDENCIES:
    -'Ialpha.py'  - located in the folder 'multi'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Tuple
import numpy as np
from proxop.multi.Ialpha import Ialpha


class Hellinger:
    r"""Compute the proximity operator and the evaluation of gamma*D.

    Where D is the  Hellinger function defined as:


                  /  ( sqrt(x) - sqrt(y) )^2   if x >= 0 and y >= 0
        D(x,y) = |
                 \  + inf                    otherwise

    'gamma' is the scale factor

     ========
     INPUTS
    ========
    x        - scalar or ND array
    y        - scalar if 'x' is a scalar , ND array with the same size as 'x' otherwise
    gamma    - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    """

    def __init__(self):
        pass

    def prox(
            self,
            x: np.ndarray,
            y: np.ndarray,
            gamma: Union[float, np.ndarray] = 1
    ) -> Tuple[np.ndarray, np.ndarray]:
        self._check(x)
        return Ialpha(alpha=2).prox(x, y, gamma)

    def __call__(self, x: np.ndarray, y: np.ndarray) -> float:
        if np.any(x < 0) or np.any(y < 0):
            return np.inf
        return np.sum((np.sqrt(x) - np.sqrt(y)) ** 2)

    def _check(self, x, gamma=1):
        if np.any(gamma <= 0):
            raise Exception("'gamma'  must be strictly positive")
        if np.size(gamma) > 1 and np.size(gamma) != np.size(x):
            ValueError(
                "'gamma' must be positive scalars or positive ND arrays" +
                " with the same size as 'x'"
            )



"""
Version : 1.0 ( 06-15-2022).

Authors  : Mbaye DIONGUE

Copyright (C) 2019

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Tuple
import numpy as np


class EpiHingeLoss:
    r"""Compute the projection and the indicator of the epigraph of phi.

    Where phi is the function defined as:
                phi(y)= w * max( y, 0)

    When the inputs are arrays, the outputs are computed element-wise
     INPUTS
    ========
     y    - ND array
     ksi  - ND array with the same size as 'y'
     w    - positive, scalar or ND array with the same size as 'ksi' [default: w=1.0]
    """

    def __init__(
            self,
            ksi: np.ndarray,
            w: Union[float, np.ndarray] = 1.0
    ):

        if np.any(w <= 0):
            raise ValueError("'w' must be positive")
        if np.size(w) > 1 and (np.size(w) != np.size(ksi)):
            raise ValueError(" 'w' must be a scalar or have the same size as 'ksi'")
        if np.size(ksi) <= 1:
            ksi = np.reshape(ksi, (-1))
        self.w = w
        self.ksi = ksi

    # proximal operator (i.e. the projection on the constraint set)
    def prox(self, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        self._check(y)
        ksi = self.ksi
        w = self.w

        if np.size(y) <= 1:
            y = np.reshape(y, (-1))

        # 3rd branch
        prox_y = np.maximum(0, y + w * ksi) / (1 + w**2)
        prox_t = w * prox_y

        # 1st and 2nd branches
        mask = np.logical_or(w * np.maximum(y, 0) <= ksi, y <= 0)
        prox_y[mask] = y[mask]
        prox_t[mask] = np.maximum(ksi[mask], 0)
        return tuple([prox_y, prox_t])

    # indicator of the constraint set
    def __call__(self, y: np.ndarray) -> float:
        if np.all(self.w * np.maximum(y, 0) <= self.ksi):
            return 0
        return np.inf

    def _check(self, y):
        if np.size(y) != np.size(self.ksi):
            raise ValueError(" 'y' must have the same size as 'ksi'")



"""
Original Version : 1.0 (01-02-2019) (updated  06-09-2022).

Author: Mbaye DIONGUE

Copyright (C) 2019

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

import numpy as np


class UnimodalSigmoid:
    r"""Computes the proximity operator and the evaluation of f.

    Where f is the 'unimodal 'sigmoid activation' function defined as

                   / (x+1/2)*log(x+1/2)+(1/2 -x)*log(1/2 -x)-(x^2+1/4)/2  if |x| < 1/2
           f(x) = | -1/4                                                |x| = 1/2
                  \ +Inf                                                otherwise


    When the input 'x' is an array, the output is computed element-wise :

    - for the function, the output is a scalar ( element-wise sum) .

    - But for the proximity operator (method 'prox'), the output has the same shape
    as the input 'x'.

     INPUTS
    ========
     x     - scalar or ND array

    ========
    Examples
    ========

     Evaluate the function  f:

     >>> UnimodalSigmoid()( 0.5 )
     -0.25

      Compute the result element-wise for vector inputs :

     >>> UnimodalSigmoid()([ 0.3, .25, .5] )
     -1.638987568156996

     Compute the proximity operator at a given point :

     >>> UnimodalSigmoid().prox(  [-2, .3, 4 ])
     array([-0.38079708,  0.07444252,  0.48201379])
    """

    def __init__(self):
        pass

    def prox(self, x: np.ndarray) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        return 1 / (1 + np.exp(-x)) - 0.5

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        prox_x = np.zeros_like(1.0 * x)
        mask = np.abs(x) < 0.5
        x_m = x[mask]
        prox_x[mask] = (
            (x_m + 0.5) * np.log(x_m + 0.5)
            + (0.5 - x_m) * np.log(0.5 - x_m)
            - 0.5 * (x_m**2 + 0.25)
        )
        prox_x[np.abs(x) == 0.5] = -0.25
        prox_x[np.abs(x) > 0.5] = np.inf
        return np.sum(prox_x)



"""
Version : 1.0 (06-18-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""


import numpy as np
from typing import Union


class CEL0:
    r"""Compute the proximity operator and the evaluation of the gamma*f.

    Where f is the continuous Exact L0 (CEL0) penalty  defined as:

    f(x)= lamb - 0.5*ai^2*(|x| - sqrt(2*lamb)/ai)^2*1_{|x| <= sqrt(2*lamb)/ai}


            'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
    x     - ND array
    gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    a     - positive, scalar or ND array with the same size as 'x'
    lamb  - positive, scalar or ND array with the same size as 'x'


     Note: When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>CEL0(lamb=2, a=3)(x) will
    return a scalar even if x is a vector:

    >>> CEL0(lamb=2, a=3)(np.array([-1, 2, 3]))
    6.0

    But as expected, >>>CEL0(lamb=2, a=3).prox(x)
    will return a vector with the same size as x:

    >>> CEL0(lamb=2, a=3).prox(np.array([-1, 2, 3]))
    array([0, 0, 3])
    """

    def __init__(
        self,
        lamb: Union[np.ndarray, float],
        a: Union[np.ndarray, float]
    ):
        if np.any(lamb <= 0):
            raise Exception("'lamb'(or all of its elements) must be strictly positive")
        if np.any(a <= 0):
            raise Exception("'a'(or all of its elements) must be strictly positive ")
        self.lamb = lamb
        self.a = a

    def prox(self, x: np.ndarray, gamma: Union[np.ndarray, float] = 1.0) -> np.ndarray:
        self._check(x, gamma)
        a = self.a
        lamb = self.lamb
        abs_x = np.abs(x)
        sign_x = np.sign(x)

        # 1st case:  a**2*gamma >=1
        mask1 = abs_x > np.sqrt(2 * gamma * lamb)
        prox_ = 1.0 * x * mask1

        if np.size(a**2 * gamma) <= 1 and (a**2 * gamma >= 1):
            return np.reshape(prox_, np.shape(x))

        # 2nd case: a**2*gamma < 1
        mask = a**2 * gamma < 1
        if np.size(lamb) > 1:
            lamb = lamb[mask]
        if np.size(a) > 1:
            a = a[mask]
        if np.size(gamma) > 1:
            gamma = gamma[mask]

        p2 = np.maximum(abs_x[mask] - np.sqrt(2 * lamb) * gamma * a, 0) / (
            1 - a**2 * gamma
        )
        prox_[mask] = sign_x[mask] * np.minimum(abs_x[mask], p2)
        return prox_

    def __call__(self, x: np.ndarray) -> float:
        self._check_size(x)
        a = self.a
        lamb = self.lamb
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))

        xx = np.abs(x) - np.sqrt(2 * lamb) / a
        mask = xx <= 0
        fun_x = lamb - 0.5 * (a * xx) ** 2 * mask
        return np.sum(fun_x)

    def _check(self, x, gamma):
        sz_x = np.size(x)
        if np.any(gamma <= 0):
            raise Exception("'gamma'(or all of its elements) must be strictly positive")
        if (np.size(gamma) > 1) and (np.size(gamma) != sz_x):
            raise Exception("'gamma' must be either scalar or the same size as 'x'")
        if (np.size(self.lamb) > 1) and (np.size(self.lamb) != sz_x):
            raise Exception("'lamb' must be either scalar or the same size as 'x'")
        if (np.size(self.a) > 1) and (np.size(self.a) != sz_x):
            raise Exception("'a' must be either scalar or the same size as 'x'")

    def _check_size(self, x):
        sz_x = np.size(x)
        if (np.size(self.lamb) > 1) and (np.size(self.lamb) != sz_x):
            raise Exception("'lamb' must be either scalar or the same size as 'x'")
        if (np.size(self.a) > 1) and (np.size(self.a) != sz_x):
            raise Exception("'a' must be either scalar or the same size as 'x'")



"""
Version : 1.0 (06-09-2022).

DEPENDENCIES:
     'newton.py' in the folder 'utils'

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np
from proxop.utils.newton import newton_


class Inverse:
    r"""Compute the proximity operator and evaluate the inverse function f.

                  / gamma* 1/x^q if x>0
           f(x)= |
                 \  +INF      otherwise with q>=1

    'gamma' is the scale factor [default: gamma=1]

    When the input 'x' is an array, the output is computed element-wise :

    -When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>Inverse()(x) will
    return a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the same
    shape as the input 'x'.So, the command >>>Inverse().prox(x)   will return an
    array with the same shape as 'x'

     INPUTS
    ========
     x     - scalar or ND array
     q     - positive >= 1, scalar or ND array with the same size as 'x' [default: q=1]
     gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]

    =======
    Examples
    ========

     Evaluate the 'direct' function :

     >>> Inverse()( 4 )
     0.25

      Compute the result element-wise for vector inputs :

     >>> Inverse(q=2, gamma=3)( [4., 2., 0.25] )
     48.9375

     Compute the proximity operator at a given point :

     >>> Inverse(2).prox(  [-2, 3, 4 ])
     array([8.85033504e-01, 3.60403299e+08, 3.60403300e+08])

     Use a scale factor 'gamma'>0 to compute the proximity operator of
     the function 'gamma*f'

     >>> Inverse(3).prox( [-2, 3, 4, np.e ], gamma=2.5 )
     array([1.23403930e+00, 4.66240534e+11, 2.38715153e+11, 4.66240534e+11])
    """

    def __init__(self, q: Union[float, np.ndarray] = 1):
        if np.size(q) > 1 and (not isinstance(q, np.ndarray)):
            q = np.array(q)
        if np.any(q < 1):
            raise ValueError("'q' must be greater than 1")
        self.q = q

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        self._check(x, gamma)
        q = self.q
        abs_x = np.abs(x)
        scale = gamma

        def fun_phi(t):
            return t ** (q + 2) - x * t ** (q + 1) - scale * q

        def der_phi(t):
            return (q + 2) * t ** (q + 1) - (q + 1) * x * t * q

        # starting point
        p_init = abs_x / 2

        # Finding the root of the polynom with the Newton method
        prox_x = newton_(fun_phi, fp=der_phi, x0=p_init, low=0, high=np.inf)

        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(x) <= 1:
            if x > 0:
                return np.sum(1 / x**self.q)
            return np.inf

        fun_x = np.zeros_like(1.0 * x)
        mask = x > 0
        q = self.q
        q_ = q
        if np.size(q) > 1:
            q_ = q[mask]
        fun_x[mask] = (1 / x[mask] ** q_)
        fun_x[np.logical_not(mask)] = np.inf
        return np.sum(fun_x)

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its elements if it is an array)"
                + "must be strictly positive"
            )
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise ValueError("gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 ( 06-21-2022).

DEPENDENCIES:
     -'prox_svd.py' - located in the folder 'utils'
     -'fun_svd.py'  - located in the folder 'utils'
     -'newton.py'  in the folder 'utils'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

import numpy as np
from proxop.utils.prox_svd import prox_svd
from proxop.utils.fun_svd import fun_svd
from proxop.utils.newton import newton_


class SchattenPenaltyLogDet:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the function defined as:

               / -log( det(X) ) +  R_p(X)^p   if X is a symmetric positive
         f(x)=|                               definite matrix
              \  + inf                        otherwise

    where

           * X = U*diag(s)*V.T \in R^{M*N}  the Singular Value decomposition of X

           * R_p(X)= ||s||_p^p  is the p-Schatten norm for matrices

           * det(X)= prod_i(s_i)    is the determinant of the matrix X

           * 'gamma*tau' is the scale factor

    Note: No checking is performed to verify whether X is symmetric or not when
    computing the proximity operator with the method 'prox'. X is assumed to be
    symmetric.

     INPUTS
    ========
     x               -  (M,N) -array_like ( representing an M*N matrix)
     p               -  scalar in ]0, 1]
     mu              - positive scalar [default: mu=1.0]
     gamma           - positive scalar  [default: gamma=1.0]
    """

    def __init__(self, p: float, mu: float = 1.0):
        if np.any(mu < 0) or np.size(mu) > 1:
            raise Exception("'mu'  must be a positive scalar")
        if np.any(p < 0) or np.any(p > 1) or np.size(p) > 1:
            raise Exception("'p'  must belong to ]0, 1] ")
        self.mu = mu
        self.p = p

    def prox(self, x: np.ndarray, gamma: float = 1.0) -> np.ndarray:
        self._check(x, gamma)
        mu = self.mu
        p = self.p

        def prox_phi(s, gam):
            # Note: same computation as in 'EpiPower.py' for q=2
            def polynom_phi(t):
                return gam * mu * p * t ** p + t ** 2 - s * t - gam

            def derive_polynom_phi(t):
                return p * gam * mu * p * t ** (p - 1) + 2 * t - s

            # starting point
            low = 1e-10
            root_init = np.abs(s)
            # Finding the root of the polynom with the Newton method
            d = newton_(
                polynom_phi, fp=derive_polynom_phi, x0=root_init, low=low, high=np.inf
            )
            return d

        return prox_svd(x, gamma, prox_phi, hermitian=True)

    def __call__(self, x: np.ndarray) -> float:
        tol = 1e-12
        p = self.p
        # Check if the matrix is symmetric
        if not np.allclose(x, np.transpose(x)):
            return np.inf

        def fun_phi(s):
            if np.any(s <= tol):
                return np.inf
            return -np.log(np.prod(s)) + self.mu * np.sum((np.abs(s)) ** p)

        return fun_svd(x, 1, fun_phi)

    def _check(self, x, gamma):
        if np.any(gamma <= 0) or np.size(gamma) > 1:
            raise Exception("'gamma'  must be a strictly positive scalar")
        if len(np.shape(x)) != 2:
            raise ValueError(
                "'x' must be an (M,N) -array_like ( representing a M*N matrix )"
            )



"""
Version : 1.0 (06-09-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class FairPotential:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is 'Fair Potential' function f defined as:

           f(x)=-  w*|x| - log(1+w|x|)

     with w > 0
    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise :

    -When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>FairPotential()(x) will
    return a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the
    same shape as the input 'x'. So, the command >>>FairPotential().prox(x)
    will return an array with the same shape as 'x'


     INPUTS
    ========
     x     - scalar or array
     w     - positive , scalar or array with the same size as 'x'
     gamma  - positive, scalar or array with the same size as 'x' [default: gamma=1]

     =======
    Examples
    ========

     Evaluate the 'direct' function :

     >>> FairPotential()( 2 )
     0.9013877113318902

      Compute the result element-wise for vector inputs :
     >>> FairPotential(gamma=2)( [-1, 0, 3, np.e] )
     6.6511571985219735

     Compute the proximity operator at a given point :

     >>> FairPotential().prox(  [-2, 3, 4, np.e] )
     array([-1.5       ,  2.5       ,  3.5       ,  2.21828183])

     Use a scale factor 'gamma'>0 to compute the proximity operator of
     the function 'gamma*f'

     >>> FairPotential().prox( [-2, 3, 4, np.e ], gamma=2.5)
     array([-0.68614066,  1.5       ,  2.38600094,  1.26147065])
    """

    def __init__(self, w: Union[float, np.ndarray] = 1):
        if np.size(w) > 1 and (not isinstance(w, np.ndarray)):
            w = np.array(w)
        if np.any(w <= 0):
            raise ValueError(
                "'w' (or all of its components if it is an array) must be strictly"
                + " positive"
            )
        self.w = w

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))

        self._check(x, gamma)
        w = self.w
        sign_x = np.sign(x)
        abs_x = np.abs(x)
        prox_x = (
            w * abs_x
            - w**2 * gamma
            - 1
            + np.sqrt((w * abs_x - w**2 * gamma) ** 2 + 4 * w * abs_x)
        )
        prox_x = sign_x * prox_x / (2 * w)
        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        return np.sum(self.w * np.abs(x) - np.log(1 + self.w * np.abs(x)))

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its elements if it is an array)"
                + "must be strictly positive"
            )
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise ValueError("gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 ( 06-14-2022).

DEPENDENCIES:
    -'newton.py'  - located in the folder 'utils'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np
from proxop.utils.newton import newton_


class L2BallBarrier:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the function defined as:

                   / -log( radius2 -  || x - c||^2    if  || x - c||_2)^2 < radius2
            f(x) =|
                  \   + inf                           otherwise


     INPUTS
    ========
     x       - ND array
     c       - ND array with the same size as 'x' ( center of the ball constraint)
     radius2 - scalar , positive  (square of the radius of the ball)
     gamma   - positive, scalar or ND array compatible with the blocks of 'x'
                 [default: gamma=1]
     axis    - None or int, axis of block-wise processing [default: axis=None]
                  axis = None --> 'x' is processed as a single vector [DEFAULT] In this
                  case 'gamma' must be a scalar.
                  axis >=0   --> 'x' is processed block-wise along the specified axis
                  (0 -> rows, 1-> columns ect. in this case, 'gamma' and 'radius'
                  must be singletons along 'axis')
    """

    def __init__(
            self,
            center: np.ndarray,
            radius2: float,
            axis=None
    ):
        if np.size(radius2) > 1 or radius2 < 0:
            raise Exception("'radius' must be a positive scalar")
        self.r = radius2
        self.c = center
        self.axis = axis

    def prox(
            self,
            x: np.ndarray,
            gamma: Union[float, np.ndarray] = 1) -> np.ndarray:

        self._check(x, gamma)
        center = self.c
        r = self.r   # square of the radius
        l2_norm = np.linalg.norm(x - center, ord=2, axis=self.axis)

        def polynom_phi(t):
            return t**3 - l2_norm * t**2 - (r + 2 * gamma) * t + r * l2_norm

        def derive_polynom_phi(t):
            return 3 * t**2 - 2 * l2_norm * t - (r + 2 * gamma)

        # starting point
        root = np.sqrt(r) / 2

        # Finding the root of the polynom with the Newton method
        root = newton_(
            polynom_phi, fp=derive_polynom_phi, x0=root, low=0, high=np.sqrt(r)
        )

        return center + (r - root**2) / (r - root**2 + 2 * gamma) * (x - center)

    def __call__(self, x: np.ndarray) -> float:
        l2_norm2 = np.sum((x - self.c) ** 2, axis=self.axis)
        if np.all(l2_norm2 < self.r):
            return -np.sum(np.log(self.r - l2_norm2))
        return np.inf

    def _check(self, x, gamma=1):
        if np.any(gamma <= 0):
            raise Exception(
                "'gamma' (or all of its components if it is an array)" +
                " must be strictly positive"
            )
        if np.size(gamma) <= 1:
            return
        if self.axis is None and np.size(gamma) > 1:
            raise ValueError(
                "'gamma' must be a scalar when the parameter 'axis' is equal to 'None'"
            )

        sz = np.shape(x)
        if len(sz) <= 1:
            self.axis = None

        if len(sz) <= 1:
            raise ValueError(
                "'gamma' must be scalar when 'x' is one dimensional"
            )

        if len(sz) > 1 and (self.axis is not None):
            sz = np.array(sz, dtype=int)
            sz[self.axis] = 1
            if np.size(gamma) > 1 and (np.prod(sz) != np.size(gamma)):
                raise ValueError(
                    "The dimension of 'gamma' is not compatible" +
                    " with the blocks of 'x'"
                )



"""
Version : 1.0 (06-08-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

import numpy as np


class GMActi:
    r"""Compute the proximity operator of the Geman MacClure activation function.

    The Geman MacClure activation function is definded as:

          / mu*arctan( sqrt( |x|/(mu - |x|) ) -  sqrt( |x|(mu - |x|) - x^2   if |x|<mu
    f(x)= | mu(pi-mu)/2                                                     if |x|=mu
          \ +INF                                                            otherwise

        with mu = 8/(3*sqrt(3))

    When the input 'x' is an array, the output is computed element-wise :

    -When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>GMActi()(x) will
    return a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the same
    shape as the input 'x'. So, the command >>>GMActi().prox(x)   will return
    an array with the same shape as 'x'.

     INPUTS
    ========
     x     - scalar or array

     =======
    Examples
    ========

     Evaluate the 'direct' function :

     >>> GMActi()( 0.5 )
      0.073622336567425

      Compute the result element-wise for vector inputs :

     >>> GMActi()( [-1, 0, 0.25] )
     0.39239604176

     Compute the proximity operator at a given point :

     >>> GMActi().prox(  [-2, 3, 4, np.e] )
     array([-1.23168057,  1.38564065,  1.44903597,  1.35607581])
    """

    def __init__(self):
        self.mu = 8 / (3 * np.sqrt(3))

    def prox(self, x: np.ndarray) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        mu = self.mu
        return mu * np.sign(x) * x**2 / (1 + x**2)

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        mu = self.mu
        abs_x = np.abs(x)

        result = mu * (np.pi - mu) / 2 * np.ones(np.shape(x))
        result[np.abs(x) > mu] = np.inf
        mask = np.abs(x) < mu
        abs_x = abs_x[mask]
        result[mask] = (
            mu * np.arctan(np.sqrt(abs_x) / (mu - abs_x))
            - np.sqrt(abs_x * (mu - abs_x))
            - x[mask] ** 2 / 2
        )
        return np.sum(result)



"""
Version : 1.0 ( 06-10-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2019

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""


import numpy as np


class Hyperplane:
    r"""Compute the projection onto the hyperplane.

    Where the hyperplane is the set defined as:

                           u.T * x = mu

     where u.T denotes the transpose of u


     INPUTS
    ========
     x      - ND array
     u      - ND array with the same size as 'x'
     mu     - scalar
    """

    def __init__(
            self,
            u: np.ndarray,
            mu: float
    ):
        self.u = u
        self.mu = mu

    def prox(self, x: np.ndarray) -> np.ndarray:
        """
        Compute the proximal operator (i.e. projection of 'x' on the hyperplane)

        Parameters
        ----------
        x : np.ndarray

        Returns
        -------
        TYPE : np.ndarray with the same size as x

        """
        self._check(x)
        coef = (self.mu - np.sum(x * self.u)) / (np.linalg.norm(self.u) ** 2)
        return x + coef * self.u

    def __call__(self, x: np.ndarray) -> float:
        """
        Indicate if the input 'x' is in the constraint set or not.

        Parameters
        ----------
        x : np.ndarray

        Returns
        -------
        0      if the input 'x' is in the set
        +inf   otherwise
        """
        self._check(x)
        tol = 1e-10
        if np.abs(np.sum(x * self.u) - self.mu) <= tol:
            return 0
        return np.inf

    def _check(self, x):
        if (np.shape(self.u)) != np.shape(x):
            raise ValueError("'u' must have the same size as 'x'")



"""
Version : 1.0 ( 06-13-2022).

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Optional
import numpy as np


class Vapnik:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the function defined as:

                        f(x) = max( ||x||_2 -epsilon, 0)


    'gamma' is the scale factor

    INPUTS
    ========
     x         - ND array
     epsilon   - positive, scalar or ND array compatible with the blocks of 'x'
     gamma     - positive, scalar or ND array compatible with the blocks of 'x'
                 [default: gamma=1]
     axis      - None or int, axis of block-wise processing [default: axis=None]
                  axis = None --> 'x' is processed as a single vector [DEFAULT] In this
                  case 'gamma' must be a scalar.
                  axis >=0   --> 'x' is processed block-wise along the specified axis
                  (0 -> rows, 1-> columns ect. In this case 'gamma' and 'epsilon' must
                  be singletons along 'axis'.
    """

    def __init__(
            self,
            epsilon: Union[float, np.ndarray],
            axis: Optional[int] = None
    ):
        if np.any(epsilon <= 0):
            raise ValueError(
                "'epsilon' (or all of its components if it is an array)" +
                " must be strictly positive"
            )
        if (axis is not None) and (axis < 0):
            axis = None
        self.epsilon = epsilon
        self.axis = axis

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1) -> np.ndarray:
        self._check(x, gamma)
        axis = self.axis
        epsilon = self.epsilon

        if np.size(x) <= 1:
            x = np.reshape(x, (-1))

        sz = np.shape(x)
        sz = np.array(sz, dtype=int)
        sz[axis] = 1
        if np.size(gamma) > 1:
            gamma = np.reshape(gamma, sz)
        if np.size(epsilon) > 1:
            epsilon = np.reshape(epsilon, sz)

        # preliminaries
        l2_x = np.linalg.norm(x, ord=2, axis=axis).reshape(sz)

        # First branch
        prox_x = 1.0 * x

        # 2nd branch
        mask_ones = np.ones(np.shape(x), dtype=bool)
        mask = mask_ones * (l2_x - epsilon > 0) * (l2_x - epsilon <= gamma)
        prox_x[mask] = (epsilon / l2_x * x)[mask]

        # 3rd branch
        mask = mask_ones * (l2_x - epsilon > gamma)
        prox_x[mask] = ((1 - gamma / l2_x) * x)[mask]

        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        l2_x = np.linalg.norm(x, ord=2, axis=self.axis)
        return np.sum(np.maximum(l2_x - self.epsilon, 0))

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its components if it is an array)" +
                " must be strictly positive"
            )
        if self.axis is None and np.size(gamma) > 1:
            raise ValueError(
                "'gamma' must be scalars when the" +
                " argument 'axis' is equal to 'None'"
            )
        if np.size(gamma) <= 1:
            return
        sz = np.shape(x)
        if len(sz) <= 1:
            self.axis = None
        if len(sz) <= 1:
            raise ValueError(
                "'gamma' must be scalar when 'x' is one dimensional"
            )
        if len(sz) > 1 and self.axis is not None:
            sz = np.array(sz, dtype=int)
            sz[self.axis] = 1
            if np.prod(sz) != np.size(gamma):
                raise ValueError(
                    "The dimension 'gamma' is not compatible" +
                    " with the blocks of 'x'"
                )



"""
Version : 1.0 ( 06-20-2022).

DEPENDENCIES:
    - 'Max.py' located in the folder 'multi'
    - 'Linf.py' located in the folder 'multi'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np
from proxop.multi.Linf import Linf


class Linf1:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is a sum of infinitive norms:

                         f(x) = \sum_{i=1}^N \sup_{1\le j \le M} |X_{i,j}|

            where X = U*diag(s)*V.T \in R^{M*N}  (Singular Value decomposition)

    'gamma' is the scale factor

     INPUTS
    ========
     x         -  (M,N) -array_like ( representing an M*N matrix )
     gamma     - positive, scalar or ND array compatible with the size of 'x'
                 [default: gamma=1]
    """

    def __init__(self):
        pass

    def prox(
            self,
            x: np.ndarray,
            gamma: Union[float, np.ndarray] = 1.0
    ) -> np.ndarray:
        return Linf(axis=1).prox(x, gamma)

    def __call__(self, x: np.ndarray) -> float:
        return Linf(axis=1)(x)



"""
Version : 1.0 ( 06-23-2022).

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class BregBoltzShannon2:
    r"""Compute the proximity operator prox_{gamma*f}^{phi}.

    Where the functions f and phi are defined as follows:


                        /  x*log(x) - w*x      if x>0
               f(x)=   |
                       \  +inf                 otherwise


                         /  u*log(u) + (1-u)*log( 1-u)    if u \in ]0, 1[
              phi(u) =  |   0                             if u \in {0,1}
                        \   +inf                          otherwise

    Note: The function phi is a Legendre type function and induces a Bregman distance

    'gamma' is the scale factor

     When the input 'x' is an array, the output is computed element-wise :

    - When calling the functions f or phi (method 'phi') the result is computed
    element-wise SUM.
    So the command >>>BregBoltzShannon2()(x) or >>>BregBoltzShannon2().phi(x) will
    return a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the
    same shape as the input 'x'.
    So, the command >>>BregBoltzShannon2.prox(x)   will return an array
    with the same shape as 'x'

    INPUTS
    ========
    x       - scalar or ND array
    w       - scalar or ND array with the same size as 'x'  [default: w=1]
    gamma   - positive, scalar or ND array with the same size as 'x'[default: gamma=1]

    =======
    Examples
    ========

    Evaluate the function f:

    >>> BregBoltzShannon2()(1)
     -1.0
     >>> BregBoltzShannon2(w=0)( [0.2, 3, 4, np.e] )
     11.237408556456117

     Evaluate the function  phi

     >> BregBoltzShannon2(w=0).phi( [-2., 0.3, 0.125, 1.] )
     inf

     Compute the proximity operator at a given point :

     >>> BregBoltzShannon2(w=2).prox(  [-2, 3, 4, np.e] )
     array([0.44986882, 0.98232606, 0.99335135, 0.97683598])
     >>> BregBoltzShannon2(2, gamma=2).prox( [-2, 3, 4] )
     array([0.44986882, 0.98232606, 0.99335135, 0.97683598])
    """

    def __init__(self, w: Union[float, np.ndarray] = 1.0):
        if np.size(w) > 1 and (not isinstance(w, np.ndarray)):
            w = np.array(w)
        self.w = w

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        self._check(x, gamma)
        w = self.w
        xx = np.exp(x + w - 1)
        prox_x = -0.5 * xx + np.sqrt(0.25 * xx**2 + xx)
        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        self._check(x, 1)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        res = np.zeros(np.shape(x))
        mask = x > 0
        w = self.w
        if np.size(w) > 1:
            w = w[mask]
        res[mask] = x[mask] * np.log(x[mask]) - w * x[mask]
        res[x <= 0] = np.inf
        return np.sum(res)

    def phi(self, u: np.ndarray) -> float:
        if np.size(u) > 1 and (not isinstance(u, np.ndarray)):
            u = np.array(u)
        if np.size(u) <= 1.0:
            u = np.reshape(u, (-1))
        res = np.zeros(np.shape(u))
        mask = (u > 0) * (u < 1)
        res[mask] = u[mask] * np.log(u[mask]) + (1 - u[mask]) * np.log(1 - u[mask])
        res[np.logical_or(u > 1, u < 0)] = np.inf
        return np.sum(res)

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError("'gamma'  must be strictly positive")
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise ValueError("gamma' must be either scalar or the same size as 'x'")
        if (np.size(self.w) > 1) and (np.size(self.w) != np.size(x)):
            raise ValueError("'w' must be either scalar or the same size as 'x'")



"""
Version : 1.0 ( 06-16-2022).

DEPENDENCIES:
     - 'EpiMax.py' located in the folder 'indicator'

Author  : Mbaye DIONGUE

Copyright (C) 2019

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Tuple
import numpy as np
from proxop.indicator.EpiMax import EpiMax


class EpiLinfNorm:
    r"""Compute the projection and the indicator of the epigraph of phi.

    Where phi is the function defined as:

                phi(y)= w * ||y||_inf

    where ||.||_inf denotes the infinity norm

     INPUTS
    ========
     y    - ND array
     ksi  - ND array compatible with the blocks of 'y' (see above)
     w    - positive, scalar or ND array with the same size as 'ksi' [default: w=1.0]
     axis - int or None, direction of block-wise processing
            When the input 'y' is an array, the computation can vary as follows:
            - axis = None --> 'y' is processed as a single vector [DEFAULT]
                                   (in this case, 'xi' must be scalar)
            - axis >= 0 --> 'y' is processed block-wise along the specified axis
            (in this case, 'xi' must be singleton along 'axis')
    """

    def __init__(
            self,
            ksi: np.ndarray,
            axis: int or None = None,
            w: Union[float, np.ndarray] = 1.0
    ):

        if np.any(w <= 0):
            raise ValueError(
                "'w' (or all of its components if it is an array) must be positive"
            )
        if np.size(w) > 1 and (np.size(w) != np.size(ksi)):
            raise ValueError(" 'w' must be a scalar or must have" +
                             " the same size as 'ksi'")
        if np.size(ksi) <= 1:
            ksi = np.array([ksi])

        self.w = w
        self.axis = axis
        self.ksi = ksi

    # proximal operator (i.e. the projection on the constraint set)
    def prox(self, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        self._check(y)
        ksi = self.ksi
        w = self.w
        axis = self.axis

        abs_y = np.abs(y)
        prox_t = EpiMax(axis=axis, ksi=ksi, w=w).prox(abs_y)[1]
        prox_t = np.maximum(0, prox_t)

        # Compute p
        prox_p = np.sign(y) * np.minimum(abs_y, prox_t / w)

        return tuple([prox_p, prox_t])

    # indicator of the constraint set
    def __call__(self, y: np.ndarray) -> float:
        self._check(y)
        tol = 1e-10
        if self.axis is None:
            y = np.reshape(y, (-1))
        mask = self.w * np.max(np.abs(y), axis=self.axis) <= self.ksi + tol
        if np.all(mask):
            return 0
        return np.inf

    def _check(self, y: np.ndarray):
        sz_ = 1
        if self.axis is not None:
            sz = np.array(np.shape(y), dtype=int)
            sz_ = np.delete(sz, self.axis)
        cond1 = self.axis is not None and np.any(sz_ != np.shape(self.ksi))
        cond2 = self.axis is None and np.any(np.size(self.ksi) != sz_)
        if cond1 or cond2:
            try:
                self.ksi = np.reshape(self.ksi, sz_)
            except ValueError:
                print("'ksi' is not compatible with the size of 'y' ")



"""
Version : 1.0 ( 06-22-2022).

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Tuple
import numpy as np
from scipy.optimize import minimize


class PersVapnik:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the function defined as:


                 /   inf_{y \in R}( |y| + 1_{[-eps*xi, eps*ksi]}    if xi >= 0
      f(y,xi) = |
                \    + inf                                         otherwise



    'gamma' is the scale factor


    INPUTS
    ========
    y          -  scalar or ND array
    xi         -  scalar or ND array with the same size as 'y'
    eps        - scalar or ND array with the same size as 'y'
    gamma      - positive, scalar or ND array compatible with the blocks of 'y'
                [default: gamma=1.0]
    """

    def __init__(
            self,
            eps: Union[float, np.ndarray],
            xi: Union[float, np.ndarray]
    ):

        if np.any(eps <= 0):
            raise Exception("'eps'  must be strictly positive")
        if np.size(eps) != np.size(xi) and np.size(eps) > 1 and np.size(xi) > 1:
            raise Exception(
                "'eps'  and 'xi' must have the same size when they are not scalars"
            )
        if np.size(xi) <= 1:
            xi = np.reshape(xi, (-1))
        if np.size(eps) <= 1:
            eps = np.reshape(eps, (-1))
        self.eps = eps
        self.xi = xi

    def prox(
            self,
            y: np.ndarray,
            gamma: Union[float, np.ndarray] = 1.0
    ) -> Tuple[np.ndarray, np.ndarray]:

        self._check(y, gamma)
        if np.size(gamma) <= 1:
            gamma = np.reshape(gamma, (-1))
        scale = gamma
        eps = self.eps
        xi = self.xi

        if np.size(y) <= 1:
            y = np.reshape(y, (-1))
        sz = np.shape(y)
        if np.size(y) > 1:
            if np.size(xi) <= 1:
                xi = xi * np.ones(sz)
            if np.size(eps) <= 1:
                eps = eps * np.ones(sz)
            if np.size(scale) <= 1:
                scale = xi * np.ones(sz)

        abs_y = np.abs(y)
        sign_y = np.sign(y)

        # 1st branch
        prox_p = np.zeros_like(1.0 * y)
        prox_q = np.zeros_like(1.0 * xi)

        # 2nd branch
        mask = (xi <= -scale * eps) * (abs_y > scale)
        prox_p[mask] = y[mask] - scale[mask] * sign_y[mask]

        # 3rd branch
        mask = (xi > -scale * eps) * (abs_y > eps * xi + scale * (1 + eps**2))
        prox_p[mask] = y[mask] - scale[mask] * sign_y[mask]
        prox_q[mask] = eps[mask] + scale[mask] * eps[mask]

        # 4th branch
        mask = (
            (abs_y > -xi / eps)
            * (abs_y > eps * xi + scale * (1 + eps**2))
            * (abs_y <= eps * xi)
        )
        prox_q[mask] = (xi[mask] + eps[mask] * sign_y[mask]) / (1 + eps[mask] ** 2)
        prox_p[mask] = eps[mask] * sign_y[mask] * prox_q[mask]

        return tuple([prox_p, prox_q])

    def __call__(self, y: np.ndarray) -> float:
        eps = self.eps
        xi = self.xi
        xi = np.reshape(xi, (-1))
        eps = np.reshape(eps, (-1))
        y = np.reshape(y, (-1))

        def fun_phi(t):
            mask = (y - t >= -eps * xi) * (y - t <= eps * xi)
            return np.sum(np.abs(y) + mask)

        res = minimize(fun_phi, y)

        return np.sum(fun_phi(res.x))

    def _check(self, y, gamma):
        if np.any(gamma <= 0):
            raise Exception("'gamma'  must be strictly positive")
        if np.size(self.xi) > 1 and np.size(self.xi) != np.size(y):
            raise Exception(
                "'xi' must be scalar or an ND array with the same size as 'y'"
            )
        if np.size(self.eps) > 1 and np.size(self.eps) != np.size(y):
            raise Exception(
                "'eps' must be scalar or an ND array with the same size as 'y'"
            )



"""
Version : 1.0 ( 06-07-2022).

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

import numpy as np


class BentIdentity:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the Bent identity function defined as:

                 /   x/2 - ln( x+ 1/2)/4     if  x > -1/2
          f(x)= |
                \  +INF                      otherwise

     When the input 'x' is an array, the output is computed element-wise :

    -When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>BentIdentity()(x) will
    return a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the same
    shape as the input 'x'.
    So, the command >>>BentIdentity().prox(x)   will return an array with
    the same shape as 'x'


     INPUTS
    ========
     x     - scalr or ND array

     ========
     Examples
     ========

    Evaluate the 'direct' function :

    >>> BentIdentity()(-0.5)
    array([inf])
    >>> BentIdentity()([-0.2, 15, 0, 0.25, 0.3])
    7.591776396181667

    Compute the proximity operator at a given point :

    >>> BentIdentity().prox(3)
    2.58113883008419
    >>> BentIdentity().prox([ -3., 0.5, 6.])
    array([-0.41886117,  0.30901699,  5.54138127])
    """

    def __init__(self):
        pass

    def prox(self, x: np.ndarray) -> np.ndarray:
        if not isinstance(x, np.ndarray):
            x = np.array(x)
        prox_x = (x + np.sqrt(x**2 + 1) - 1) / 2
        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        if not isinstance(x, np.ndarray):
            x = np.array(x)
        if np.any(x <= -0.5):
            return np.inf
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        result = x / 2 - np.log(x + 1 / 2) / 4
        return np.sum(result)



"""
Version : 1.0 (06-18-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Tuple
import numpy as np


class ConicL0:
    r"""Compute the proximity operator and the evaluation of the gamma*f.

    where the function f is defined as:

                   f(x,d) = gamma \ell_0(x) + \iota_{S}(x,d)

    with S = {x \in C^N, d \in C^N, s.t. (\forall n \in {1,...,N})
              \exist delta_n \in [-Delta_n,Delta_n] s.t. d_n = \delta_n x_n
              }
    where \Delta \in [0,+\infty)^N

    When the inputs '(x,d)' are arrays, the outputs '(p,q)' are computed element-wise.

     INPUTS
    ========
     x     - ND array complex valued
     delta - real positive,  scalar or ND array with the same size as 'x'
     d     - ND array complex valued with the same size as 'x'
     gamma - positive, scalar

    Note: When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>ConicL0(Delta=2, d=2*x)(x) will
    return a scalar even if x is a vector:
    >>> x=np.array([ 1+1j, 4+2J])
    >>> x
    array([1.+1.j, 4.+2.j])
    >>> ConicL0(delta=3, d=2*x)(x)
    2

    But as expected, >>> ConicL0(Delta=3, d=2*x).prox(x)
    will return two vectors with the same size as x:

    >>> px=ConicL0(delta=3, d=2*x).prox(x)
    [array([1.+1.j, 4.+2.j]), array([2.+2.j, 8.+4.j])]
    """

    def __init__(
        self,
        delta: Union[np.ndarray, float],
        d: np.ndarray
    ):

        if np.any(delta < 0):
            raise Exception("'delta'(or all of its elements) must be positive or null")
        self.delta = delta
        self.d = d

    def prox(self, x: np.ndarray, gamma: float = 1.0) -> Tuple[np.ndarray, np.ndarray]:
        self._check(x, gamma)
        self._check_size(x)
        delta = self.delta
        d = self.d

        prox_x = np.zeros_like(1.0 * x)
        prox_q = np.zeros_like(1.0 * x)

        z = np.real(x * np.conj(d))
        abs_d2 = np.abs(d) ** 2
        abs_x2 = np.abs(x) ** 2

        eta = np.sqrt((abs_d2 - abs_x2) ** 2 + 4 * z ** 2)

        delta_hat = delta * np.ones(np.shape(x))
        mask = z != 0
        delta_hat1 = (eta + abs_d2 - abs_x2) / (2 * np.abs(z))
        delta_hat[mask] = (np.minimum(delta_hat1, delta) * np.sign(z))[mask]
        delta_hat[(z == 0) * (abs_x2 >= abs_d2)] = 0

        seuil_mod = (np.abs(delta_hat * x - d) ** 2) / (1 + delta_hat**2) + 2 * gamma
        ind_seuil = abs_x2 + abs_d2 >= seuil_mod

        upmod = (x + delta_hat * d) / (1 + delta_hat**2)

        prox_x[ind_seuil] = upmod[ind_seuil]
        prox_q[ind_seuil] = upmod[ind_seuil] * delta_hat[ind_seuil]
        return tuple([prox_x, prox_q])

    def __call__(self, x: np.ndarray) -> float:
        self._check_size(x)
        d = self.d
        delta = self.delta
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        if np.size(d) <= 1:
            d = np.reshape(d, (-1))
        tol = 1e-15
        # check the constraints
        # x and d does not have the same arg
        if abs(np.sum(np.imag(x * np.conj(d)))) > tol:
            return np.inf

        # x and d have the same arg
        mask_x = x == 0
        mask_d = d == 0
        if np.any(mask_x != mask_d):
            return np.inf
        # evaluate the l0 function
        fun_x = np.sum(x != 0)

        if np.size(mask_x) == np.size(x):
            return fun_x

        mask_x = x != 0
        # delta not in [-Delta;Delta]
        if np.any(np.abs(x[mask_x]) / np.abs(d[mask_x]) > delta):
            return np.inf

        return fun_x

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise Exception("'gamma' must be strictly positive")
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise Exception("'gamma' must be either scalar or the same size as 'x'")

    def _check_size(self, x):
        sz_x = np.size(x)
        if np.size(self.d) > 1 and (np.size(self.d) != sz_x):
            raise Exception("'d' must have the same size as 'x'")
        if (np.size(self.delta) > 1) and (np.size(self.delta) != sz_x):
            raise Exception("'delta' must be either scalar or the same size as 'x'")



"""
Version : 1.0 (06-19-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Optional
import numpy as np


class TruncatedNorm:
    r"""Compute the proximity operator and the evaluation of the gamma*f.

    where the function f is defined as:

         f(x)= min{ ||x||_2^2, w}

     'gamma' is the gamma factor

     INPUTS
     ========
     x     - ND array
     gamma - positive, scalar [default: gamma=1]
     w     - positive, scalar or ND array compatible with the blocks of 'x'
     axis  - None or int, axis of block-wise processing [default: axis=None]
              axis = None --> 'x' is processed as a single vector [DEFAULT] In this
              case 'gamma' must be a scalar.
              axis >=0   --> 'x' is processed block-wise along the specified axis
              (0 -> rows, 1-> columns ect. In this case, 'gamma' must be singleton
              along 'axis'.
    """

    def __init__(
        self,
        w: Union[float, np.ndarray],
        axis: Optional[int] = None
    ):

        if np.any(w <= 0):
            raise ValueError("'w'  must be strictly positive")
        self.axis = axis
        self.w = w

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        self._check(x, gamma)
        w = self.w
        sz0 = np.shape(x)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        sz = np.shape(x)
        axis = self.axis
        sz = np.array(sz, dtype=int)
        sz[axis] = 1

        if np.size(w) > 1:
            w = np.reshape(w, sz)
        if np.size(gamma) > 1:
            gamma = np.reshape(gamma, sz)
        mask = np.sum(x**2, axis=axis).reshape(sz) < w * (1 + 2 * gamma)
        prox_x = x / (1 + 2 * gamma * mask)
        prox_x = np.reshape(prox_x, sz0)
        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        res = np.minimum(np.sum(x**2, axis=self.axis), self.w)
        return np.sum(res)

    def _check(self, x, gamma):
        sz = np.shape(x)
        if np.any(gamma <= 0):
            raise ValueError("'gamma'  must be strictly positive")
        if len(sz) <= 1:
            self.axis = None
        if len(sz) <= 1 and (np.size(gamma) > 1 or np.size(self.w) > 1):
            raise ValueError(
                "'gamma' and 'w' must be scalars when 'x' is one dimensional"
            )
        if len(sz) > 1 and self.axis is not None:
            sz = np.array(sz, dtype=int)
            sz[self.axis] = 1
            if 1 < np.size(gamma) != np.prod(sz):
                raise ValueError(
                    "The dimension of 'gamma' is not"
                    + " compatible with the blocks of 'x'"
                )
            if np.size(self.w) > 1 and (np.prod(sz) != np.size(self.w)):
                raise ValueError(
                    "The dimension of 'w' is not compatible with the blocks of 'x'"
                )



"""
Evaluate phi(s) where s is the vector that is composed by singular values of x.

Version : 1.0 (06-14-2022).

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

import numpy as np


def fun_svd(
        x: np.ndarray,
        gamma: float or np.ndarray,
        fun_phi, hermitian: bool = False
) -> np.ndarray:
    r"""Evaluate phi(s) where s is the vector that is composed by singular values of x.

                        f(x) = gamma * phi(s)

     Where X = U*diag(s)*V.T \in R^{M*N}  is the Singular Value decomposition of X

      INPUTS
     ========
    x       - ND array
    gamma   - positive, scalar or ND array compatible with the size of 'x'
    fun_phi - function handle with one argument at least
    """
    # spectral decomposition
    u, s, vh = np.linalg.svd(x, full_matrices=True, hermitian=hermitian)

    return gamma * fun_phi(s)



"""
Version : 1.0 (06-10-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

import numpy as np


class SmoothedFermiDiracEntropy:
    r"""Computes the proximity operator and the evaluation of f.

    Where f is the 'Smoothed Fermi-Dirac Entropy' function defined as:

                      /  xlog(x)+ (1-x)log(1-x) -0.5*x^2        if 0< x < 1
               f(x)= |   0                                      if x in {0, 1}
                     \  +INF                                    otherwise


    When the input 'x' is an array, the output is computed element-wise :

    -When calling the function, the output is a scalar (sum of the
    element-wise results).

    - But for the proximity operator (method 'prox'), the output has the
    same shape as the input 'x'.

     INPUTS
    ========
     x     - scalar or ND array

    =======
    Examples
    ========

     Evaluate the function  f:

     >>> SmoothedFermiDiracEntropy()( 0.5 )
     -0.8181471805599453

      Compute the result element-wise sum for vector inputs :

     >>> SmoothedFermiDiracEntropy()( [1, .3, 0.5] )
     -1.4740114826148387

     Compute the proximity operator at a given point :
     >> SmoothedFermiDiracEntropy().prox(   3 )
     0.9525741268224334
     >>> SmoothedFermiDiracEntropy().prox(  [-2, 3, 4 ])
     array([0.11920292, 0.95257413, 0.98201379])
    """

    def __init__(self):
        pass

    def prox(self, x: np.ndarray) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        return 1 / (1 + np.exp(-x))

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        fun_x = np.zeros_like(1.0 * x)
        mask = np.logical_and(x > 0, x < 1)
        fun_x[mask] = (
            x[mask] * np.log(x[mask])
            + (1 - x[mask]) * np.log(1 - x[mask])
            - 0.5 * x[mask] ** 2
        )
        mask2 = np.logical_or(x < 0, x > 1)
        fun_x[mask2] = np.inf

        return np.sum(fun_x)



"""
Version : 1.0 ( 06-10-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2019

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class BoxConstraint:
    r"""Compute the projection onto the box constraint set.

                           low <= x <= high

     INPUTS
    ========
     x    - ND array
     low  - scalar or ND array with the same size as 'x'
     high - scalar or ND array with the same size as 'x'
    """

    def __init__(
            self,
            low: Union[float, np.ndarray],
            high: Union[float, np.ndarray]
    ):
        if np.any(low > high):
            raise Exception("'low' must be lower than 'high'")
        self.low = low
        self.high = high

    # proximal operator (i.e. projection on the box)
    def prox(self, x: np.ndarray) -> np.ndarray:
        self._check(x)
        return np.maximum(self.low, np.minimum(x, self.high))

    # indicator of the box
    def __call__(self, x: np.ndarray) -> float:
        """Indicate if the input 'x' is in the box or not.

        Parameters
        ----------
        x : np.ndarray

        Returns
        -------
        0      if the input 'x' is in the box
        +inf   otherwise
        """
        self._check(x)
        if np.all(self.low <= x) and np.all(x <= self.high):
            return 0
        return np.inf

    def _check(self, x: np.ndarray):
        if (np.size(self.low) > 1) and (np.shape(self.low) != np.shape(x)):
            raise Exception("'low' must be either scalar or the same size as 'x'")
        if (np.size(self.high) > 1) and (np.shape(self.high) != np.shape(x)):
            raise Exception("'high' must be either scalar or the same size as 'x'")



"""
Version : 1.0 ( 06-16-2022).

DEPENDENCIES:
     - 'Simplex.py' located in the folder 'indicator'

Author  : Mbaye DIONGUE

Copyright (C) 2019

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np
from proxop.indicator.Simplex import Simplex


class ExtendedSimplex(Simplex):
    r"""Compute the projection onto the extended simplex.

    Where the extended simplex is the set defined as:

       x >= 0 and  (1,..., 1).T * X <= eta


     where (1, ..., 1) is a ND array with all components equal to one,
     and (1,..., 1).T denotes its transpose

     INPUTS
    ========
     x    - ND array
     eta  - positive, scalar or ND array compatible with the blocks of 'x'
    """

    def __init__(
            self,
            eta: Union[float, np.ndarray]
    ):
        super().__init__(eta, axis=None)

    # proximal operator (i.e. projection on the extended simplex)
    def prox(self, x: np.ndarray) -> np.ndarray:

        # pre-project onto the positive quarter
        prox_x = np.maximum(x, 0)
        scalar_prod = np.sum(prox_x)
        if scalar_prod <= self.eta:
            return prox_x
        return Simplex(self.eta).prox(x)

    def __call__(self, x: np.ndarray) -> float:
        """
        Indicate if the input 'x' is in the constraint set or not.

        Parameters
        ----------
        x : np.ndarray

        Returns
        -------
        0      if the input 'x' is in the set
        +inf   otherwise
        """

        scalar_prod = np.sum(x)
        if np.all(x >= 0) and (scalar_prod <= self.eta):
            return 0
        return np.inf



"""
Version : 1.0 ( 06-21-2022).

DEPENDENCIES:
     - 'prox_svd.py' located in the folder 'utils'
     - 'prox_svd.py' located in the folder 'utils'
     - 'lambert_W.py' located in the folder 'utils'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

import numpy as np
from proxop.utils.prox_svd import prox_svd
from proxop.utils.fun_svd import fun_svd
from proxop.utils.lambert_W import lambert_W


class NeumannEntropy:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the Neumann's entropy function defined as:

               /  trace( Xlog(X))    if X is a symmetric positive matrix
         f(x)=|
              \  + inf                          otherwise

    where

           * X = U*diag(s)*V.T \in R^{M*N}  the Singular Value decomposition of X

          * tr(X)= \sum_{i=0}^{N-1}(s_i)   is the trace of the matrix X

          * 'gamma' is the scale factor

     Note:
        No checking is performed to verify whether X is symmetric or not when computing
        the proximity operator with the method 'prox'. X is assumed to be symmetric.

     INPUTS
    ========
     x               -  (M,N) -array_like ( representing an M*N matrix)
     gamma           - positive scalar  [default: gamma=1]
    """

    def __init__(self):
        pass

    def prox(self, x: np.ndarray, gamma: float = 1) -> np.ndarray:
        self._check(x, gamma)

        def prox_phi(s, gam):
            phi_s = np.zeros_like(s)
            # we use a taylor approximation of the LambertW function to avoid
            # divergence when the input is "too big"
            u = np.log(gam) + s / gam - 1
            mask = u > 100

            phi_s[mask] = u[mask] - u[mask] / (1 + u[mask]) * np.log(u[mask])

            # we use the Lambert_W function otherwise
            mask = np.logical_not(mask)
            if np.size(gam) > 1:
                gam = gam[mask]
            phi_s[mask] = gam * lambert_W(gam * np.exp(s[mask] / gam - 1))

            return phi_s

        return prox_svd(x, gamma, prox_phi, hermitian=True)

    def __call__(self, x: np.ndarray) -> float:
        self._check(x)
        tol = 1e-10
        # Check if the matrix is symmetric
        if not np.allclose(x, np.transpose(x)):
            return np.inf

        def fun_phi(s):
            if np.any(s <= tol):
                return np.inf
            return np.sum(s * np.log(s))

        return fun_svd(x, 1, fun_phi)

    def _check(self, x, gamma=1):
        if np.any(gamma <= 0) or np.size(gamma) > 1:
            raise Exception("'gamma'  must be a strictly positive scalar")
        if len(np.shape(x)) != 2:
            raise ValueError(
                "'x' must be an (M,N) -array_like ( representing a M*N matrix )"
            )



"""
Version : 1.0 (06-09-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np
from proxop.utils.newton import newton_


class LogPower:
    r"""Compute the proximity operator and the evaluation of gamma*f.

     Where f is the 'logarithm+power' function defined as:

                       / ( - log(x) + w * x^ q )     if  x> 0
                f(x)= |
                      \  +INF                                 otherwise

         with w > 0 and q >= 1

     'gamma*tau' is the scale factor

     When the input 'x' is an array, the output is computed element-wise :

     -When calling the function, the output is a scalar (sum of the element-wise
    results ) .

     - But for the proximity operator (method 'prox'), the output has the same shape
     as the input 'x'.

      INPUTS
     ========
      x     - ND array
      q     - positive >=1, scalar or ND array with the same size as 'x'
      w     - positive , scalar or ND array with the same size as 'x'
      gamma - positive, scalar or ND array with the same size as 'x' (default: gamma=1)

     =======
     Examples
     ========

      Evaluate the function  f:

      >>> LogPower(q=1)( 2. )
      1.718281828459045

       Compute the result element-wise for vector inputs :

      >>> LogPower(q=2, w=[1, 2, 3])( [1, 3, np.e] )
      39.068556008123835

      Compute the proximity operator at a given point :

      >>> LogPower(2, [1,2,3]).prox(  [-2, 3, 4 ])
      array([0.33333333, 0.83851648, 0.75951783])

      Use a scale factor 'gamma'>0 to compute the proximity operator of
      the function 'gamma*f'

      >>> LogPower(2, [1,2,3, 3]).prox( [-2, 3, 4, np.e ] , gamma=2.5 )
      array([0.5       , 0.63221419, 0.5395781 , 0.48925544])
    """

    def __init__(
        self,
        q: Union[float, np.ndarray],
        w: Union[float, np.ndarray] = 1.0
    ):
        if np.size(w) > 1 and (not isinstance(w, np.ndarray)):
            w = np.array(w)
        if np.size(q) > 1 and (not isinstance(q, np.ndarray)):
            q = np.array(q)
        if np.any(w < 0):
            raise ValueError(
                "'w (or all its components if it is an array) must"
                + " be strictly positive"
            )
        self.w = w
        if np.any(q < 1):
            raise ValueError(
                "'q' (or all its components if it is an array) must be greater than 1"
            )
        self.q = q

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        self._check(x, gamma)
        q = self.q
        w = self.w

        def polynom_phi(t):
            return q * w * gamma * t**q + t**2 - x * t - gamma

        def derive_polynom_phi(t):
            return q * q * w * gamma * t ** (q - 1) + 2 * t - x

        abs_x = np.abs(x)
        prox_x = abs_x / 2

        prox_x = newton_(
            polynom_phi, fp=derive_polynom_phi, x0=prox_x, low=1e-16, high=np.inf
        )

        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        if (np.size(self.w) > 1) and (np.shape(self.w) != np.shape(x)):
            raise ValueError("' w' must be either scalar or the same size as 'x'")

        result = np.zeros_like(1.0 * x)
        mask = x > 0
        ww = self.w
        qq = self.q
        if np.size(self.w) > 1:
            ww = self.w[mask]
        if np.size(self.q) > 1:
            qq = self.q[mask]
        result[mask] = -np.log(x[mask]) + ww * x[mask] ** qq
        result[x <= 0] = np.inf
        return np.sum(result)

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its components if it is an array) must "
                + "be strictly positive"
            )
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise ValueError("gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 ( 06-13-2022).

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""
from typing import Union, Optional
import numpy as np


class L2Positive:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the euclidian positive norm:

                          /  ||x||_2        if x>=0
                f(x) =   |
                         \  +inf            otherwise


     'gamma' is the scale factor

     INPUTS
    ========
    x         - ND array
    gamma    - positive, scalar or ND array compatible with the blocks of 'x'
                 [default: gamma=1]
    axis     - None or int, axis of block-wise processing [default: axis=None]
                  axis = None --> 'x' is processed as a single vector [DEFAULT] In this
                  case 'gamma' must be a scalar.
                  axis >=0   --> 'x' is processed block-wise along the specified axis
                  (0 -> rows, 1-> columns ect. In this case, 'gamma' must be singleton
                  along 'axis'.
    """

    def __init__(
            self,
            axis: Optional[int] = None
    ):
        self.axis = axis

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        self._check(x, gamma)
        x_plus = np.maximum(x, 0)
        l2_x2 = np.sqrt(np.sum(x_plus ** 2, axis=self.axis, keepdims=True))
        l2_x2 = np.maximum(0, 1 - gamma / l2_x2)
        prox_x = x_plus * l2_x2
        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        x[x < 0] = np.inf
        l2_x_pos = np.sqrt(np.sum(x ** 2, axis=self.axis))
        return np.sum(l2_x_pos)

    def _check(self, x, gamma=1):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its components if it is an array)" +
                " must be strictly positive"
            )
        if self.axis is None and np.size(gamma) > 1:
            raise ValueError(
                "'gamma' must be a scalar when the parameter 'axis' is equal to 'None'"
            )
        if np.size(gamma) <= 1:
            return
        sz = np.shape(x)
        if len(sz) <= 1:
            self.axis = None
        if len(sz) <= 1:
            raise ValueError(
                "'gamma' must be scalar when 'x' is one dimensional"
            )
        if len(sz) > 1 and (self.axis is not None):
            sz = np.array(sz, dtype=int)
            sz[self.axis] = 1
            if np.size(gamma) > 1 and (np.prod(sz) != np.size(gamma)):
                raise ValueError(
                    "The dimension of 'gamma' is not compatible" +
                    " with the blocks of 'x'"
                )



"""
Version : 1.0 ( 06-14-2022).

DEPENDENCIES:
     - 'L1Ball.py' - located in the folder 'indicator'
     - 'Max.py' in the folder 'multi'

Author  : Mbaye DIONGUE

Copyright (C) 2019

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""


import numpy as np
from proxop.indicator.L1Ball import L1Ball


class L1_2Ball:
    """Compute the projection and the indicator of the constraint set S.

    Where every vector X belonging to S verifies:

         ||x^(1)||_2 + ... + ||x^(B)||_2 <= eta


     INPUTS
    ========
     x         - ND array
     eta       - positive scalar
     axis - int or None, direction of block-wise processing
            When the input 'x' is an array, the computation can vary as follows:
            - axis = None --> 'x' is processed as a single vector [DEFAULT]
            - axis >= 0 --> 'x' is processed block-wise along the specified axis
               (axis=0 -> rows, axis=1 -> columns etc.).
    """

    def __init__(
            self,
            eta: float,
            axis: int or None = None
    ):
        if np.size(eta) > 1 or eta <= 0:
            raise ValueError("'eta' must be a positive scalar")
        self.eta = eta
        self.gamma = 1.0
        self.axis = axis

    def prox(self, x):
        xa = np.linalg.norm(x, ord=2, axis=self.axis, keepdims=True)
        # compute the projection
        if np.sum(xa) <= self.eta:
            # the vector if already in the set => it is equal to its projection
            return x

        # project the block norms
        beta = L1Ball(eta=self.eta, axis=self.axis).prox(xa)

        # compute the scale factors
        ya = beta / xa
        ya[xa == 0] = 0

        # rescale the blocks
        prox_x = ya * x
        return prox_x

    def __call__(self, x) -> float:
        """Indicate if the input 'x' is in the constraint set or not.

        Parameters
        ----------
        x : np.ndarray

        Returns
        -------
        0      if the input 'x' is in the set
        +inf   otherwise
        """
        l2_norm_x = np.linalg.norm(x, ord=2, axis=self.axis, keepdims=True)
        if np.sum(l2_norm_x) <= self.eta:
            return 0
        return np.inf



"""
Version : 1.0 (06-09-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

import numpy as np


class ISRU:
    r"""Compute the proximity operator of the Inverse Square Root Unit function.

                   / -(x^2)/2-sqrt(1-x^2)   if |x| <= 1
           f(x) = |
                  \ +Inf                     otherwise


    When the input 'x' is an array, the output is computed element-wise :

    -When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>ISRU()(x) will
    return a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the same shape
    as the input 'x'.
    So, the command >>>ISRU().prox(x)   will return an array with the same shape as 'x'

     INPUTS
    ========
     x     - scalar or ND array

     =======
     Examples
     ========

     >>>ISRU()( -3 )
     INF
     >>> ISRU()( [1, -0.25, -0.5] )
     -2.490521240336293

     Compute the proximity operator at a given point :

     >>>ISRU().prox(-3)
     -0.9486832980505138
     >>>ISRU().prox(  [-2, 3, -0.5] )
      array([-0.89442719,  0.9486833 , -0.4472136 ])
    """

    def __init__(self):
        pass

    def prox(self, x: np.ndarray) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        return x / np.sqrt(1 + x**2)

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))

        result = np.zeros(np.shape(x))
        mask = np.abs(x) <= 1
        result[mask] = -x[mask] ** 2 / 2 - np.sqrt(1 - x[mask] ** 2)
        result[np.abs(x) > 1] = np.inf
        return np.sum(result)



"""
Version : 1.0 ( 06-08-2022).

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

import numpy as np


class ElliotActi:
    r"""Compute the proximity operator and the evaluation of f.

    Where f is the Elliot Activation function:

                   / -|x|-ln(1-|x|)-(x^2)/2  if |x| < 1
           f(x) = |
                  \ +Inf                     otherwise


    When the input 'x' is an array, the output is computed element-wise :

    -When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>ElliotActi(x) will
    return a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the same
    shape as the input 'x'.So, the command >>>ElliotActi.prox(x)
    will return an array with the same shape as 'x'

     INPUTS
    ========
     x     - scalar or ND array

    ========
    Examples
    ========

    Evaluate the 'direct' function :
    >>> ElliotActi()(-0.5)
    0.06814718055994529
    >>>ElliotActi()([-0.2, -0.5, 0, 0.25, 0.3])
    0.08939774826466834

    Compute the proximity operator at a given point :
    >>> ElliotActi().prox(3)
    0.75
    >>> ElliotActi().prox([ -3., 0.5, 6.])
    array([-0.75      ,  0.33333333,  0.85714286])
    """

    def __init__(self):
        pass

    def prox(self, x: np.ndarray) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        return x / (1 + np.abs(x))

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))

        prox_x = np.zeros_like(x)
        mask = np.abs(x) < 1
        prox_x[mask] = (-np.abs(x[mask]) - np.log(1 - np.abs(x[mask]))
                        - 0.5 * x[mask] ** 2)
        prox_x[np.abs(x) >= 1] = np.inf
        return np.sum(prox_x)



"""
Version : 1.0 ( 06-15-2022).

DEPENDENCIES:
     - 'newton.py' located in the folder 'utils'

Author  : Mbaye DIONGUE

Copyright (C) 2019

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Tuple
import numpy as np
from proxop.utils.newton import newton_


class EpiLog:
    r"""Compute the projection and the indicator of the epigraph of phi.

    Where phi is the function defined as:

                          / -w*log(y)    if y>0
                phi(y)=  |
                         \ + inf        otherwise

    When the inputs are arrays, the outputs are computed element-wise

     INPUTS
    ========
     y    - ND array
     ksi  - ND array with the same size as 'y'
     w    - positive, scalar or ND array with the same size as 'ksi' [default: w=1.0]
    """

    def __init__(
            self,
            ksi: np.ndarray,
            w: Union[float, np.ndarray] = 1.0
    ):

        if np.any(w <= 0):
            raise ValueError(
                "'w' (or all of its components if it is an array) must be positive"
            )
        if np.size(w) > 1 and (np.size(w) != np.size(ksi)):
            raise ValueError(" 'w' must be a scalar or must have" +
                             " the same size as 'ksi'")
        if np.size(ksi) <= 1:
            ksi = np.array([ksi])
        self.w = w
        self.ksi = ksi

    # proximal operator (i.e. the projection on the constraint set)
    def prox(self, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        self._check(y)
        ksi = self.ksi
        w = self.w

        # Newton's method .
        eps = 1e-20

        def fun_phi(t):
            pp = 0.5 * (y + np.sqrt(y**2 + 4 * w * t))
            return -w * np.log(pp + eps) - ksi - t

        def der_phi(t):
            dp = w / np.sqrt(y**2 + 4 * w * t)
            pp = 0.5 * (y + np.sqrt(y**2 + 4 * w * t))
            return -w * dp / (pp + eps) - 1

        # starting point
        low = 0
        root = np.abs(y)

        # Finding the root of the polynom with the Newton method
        root = newton_(fun_phi, fp=der_phi, x0=root, low=low, high=np.inf)
        prox_p = 0.5 * (y + np.sqrt(y**2 + 4 * w * root))
        prox_t = ksi + root

        return tuple([prox_p, prox_t])

    # indicator of the constraint set
    def __call__(self, y: np.ndarray) -> float:
        if np.all(y > 0) and np.all(-self.w * np.log(y) <= self.ksi):
            return 0
        return np.inf

    def _check(self, y):
        if np.size(y) != np.size(self.ksi):
            raise ValueError(" 'y' must have the same size as 'ksi'")
